from onetick.py.core.source import Source, _Source  # _Source for backward compatibility

from .ticks import Tick, Ticks, TTicks
from .data_source import DataSource, Custom

from .cache import ReadCache
from .csv import CSV, LocalCSVTicks
from .data_file import DataFile
from .empty import Empty
from .custom import Orders, Quotes, Trades, NBBO
from .order_book import ObSnapshot, ObSnapshotFlat, ObSnapshotWide, ObSummary, ObSize, ObVwap, ObNumLevels
from .odbc import ODBC
from .parquet import ReadParquet
from .query import query, Query
from .snapshots import ReadSnapshot, ShowSnapshotList, FindSnapshotSymbols
from .split_query_output_by_symbol import SplitQueryOutputBySymbol, by_symbol
from .symbology_mapping import SymbologyMapping
from .symbols import Symbols
from .pit import PointInTime



import warnings

import onetick.py.types as ott
from onetick.py.core._source._symbol_param import _SymbolParamColumn


globals().setdefault("__warningregistry__", {})  # otherwise doctests fails


class SymbolType:
    def __init__(self):
        """
        You can get symbol name and symbol parameters with this class.

        Examples
        --------
        >>> symbols = otp.Symbols('SOME_DB')
        >>> symbols['PARAM'] = 'PAM'
        >>> ticks = otp.DataSource('SOME_DB', tick_type='TT')
        >>> ticks['SYMBOL_PARAM'] = ticks.Symbol['PARAM', str]
        >>> ticks['SYMBOL_NAME'] = ticks.Symbol.name
        >>> ticks = otp.merge([ticks], symbols=symbols)
        >>> otp.run(ticks)
                             Time  X SYMBOL_PARAM SYMBOL_NAME
        0 2003-12-01 00:00:00.000  1          PAM          S1
        1 2003-12-01 00:00:00.000 -3          PAM          S2
        2 2003-12-01 00:00:00.001  2          PAM          S1
        3 2003-12-01 00:00:00.001 -2          PAM          S2
        4 2003-12-01 00:00:00.002  3          PAM          S1
        5 2003-12-01 00:00:00.002 -1          PAM          S2

        See also
        --------
        | :ref:`Databases, symbols, and tick types <symbols_concept>`
        | :ref:`api/misc/symbol_param:Symbol Parameters Objects`
        """

        self._name = _SymbolParamColumn("_SYMBOL_NAME", str)

    @property
    def name(self):
        """
        Get symbol name.

        Returns
        -------
        _SymbolParamColumn

        Examples
        --------
        >>> symbols = otp.Symbols('SOME_DB')
        >>> ticks = otp.DataSource('SOME_DB', tick_type='TT')
        >>> ticks['SYMBOL_NAME'] = ticks.Symbol.name
        >>> ticks = otp.merge([ticks], symbols=symbols)
        >>> otp.run(ticks)
                             Time  X SYMBOL_NAME
        0 2003-12-01 00:00:00.000  1          S1
        1 2003-12-01 00:00:00.000 -3          S2
        2 2003-12-01 00:00:00.001  2          S1
        3 2003-12-01 00:00:00.001 -2          S2
        4 2003-12-01 00:00:00.002  3          S1
        5 2003-12-01 00:00:00.002 -1          S2
        """
        return self._name

    def get(self, name, dtype, default=None):
        """
        Get symbol parameter by name.

        Parameters
        ----------
        name: str
            The name of the symbol parameter to retrieve.
        dtype: type
            The expected data type of the symbol parameter value.
        default: Any, optional
            The default value to return if the symbol parameter is not found.
            Default is ``None``, which means that default value of dtype will be used.

        Returns
        -------
        Operation

        Examples
        --------
        >>> symbols = otp.Symbols('SOME_DB')
        >>> symbols['PARAM'] = 'PAM'
        >>> ticks = otp.DataSource('SOME_DB', tick_type='TT')
        >>> ticks['SYMBOL_PARAM'] = ticks.Symbol.get(name='PARAM', dtype=str, default='default')
        >>> ticks = otp.merge([ticks], symbols=symbols)
        >>> otp.run(ticks)
                             Time  X SYMBOL_PARAM
        0 2003-12-01 00:00:00.000  1          PAM
        1 2003-12-01 00:00:00.000 -3          PAM
        2 2003-12-01 00:00:00.001  2          PAM
        3 2003-12-01 00:00:00.001 -2          PAM
        4 2003-12-01 00:00:00.002  3          PAM
        5 2003-12-01 00:00:00.002 -1          PAM

        >>> symbol1 = otq.Symbol('SOME_DB::S1', params={'PARAM': 1})
        >>> symbol2 = otq.Symbol('SOME_DB::S2')
        >>> data = otp.DataSource(tick_type='TT')
        >>> data['P'] = data.Symbol.get(name='PARAM', dtype=int, default=10)
        >>> data = otp.merge([data], symbols=[symbol1, symbol2], identify_input_ts=True)
        >>> otp.run(data)
                             Time   P  X  SYMBOL_NAME TICK_TYPE
        0 2003-12-01 00:00:00.000   1  1  SOME_DB::S1        TT
        1 2003-12-01 00:00:00.000  10 -3  SOME_DB::S2        TT
        2 2003-12-01 00:00:00.001   1  2  SOME_DB::S1        TT
        3 2003-12-01 00:00:00.001  10 -2  SOME_DB::S2        TT
        4 2003-12-01 00:00:00.002   1  3  SOME_DB::S1        TT
        5 2003-12-01 00:00:00.002  10 -1  SOME_DB::S2        TT

        """
        if default is None:
            default = ott.default_by_type(dtype)
        return self._get(item=name, dtype=dtype, default=default)

    @staticmethod
    def _get(item, dtype=str, default=None):
        return _SymbolParamColumn(item, dtype, default=default)

    def __getattr__(self, item):
        """
        Get symbol parameter by name. Notice, that symbol parameter type will be string.

        .. deprecated:: 1.74.0
           Please, use :py:meth:`~onetick.py.core._source.symbol.__getitem__` method.

        Returns
        -------
        _SymbolParamColumn
        """
        if not item.startswith('_') and item != 'pytest_mock_example_attribute_that_shouldnt_exist':
            warnings.warn("`__getattr__` method is deprecated. Please, use `__getitem__` method instead.",
                          FutureWarning, stacklevel=2)
        return self._get(item)

    def __getitem__(self, item):
        """
        Get symbol parameter by name.

        Parameters
        ----------
        item: tuple
            The first parameter is string - symbol parameter name. The second one is symbol parameter type.

        Returns
        -------
        _SymbolParamColumn

        Examples
        --------
        The second parameter is symbol parameter's type.

        >>> symbols = otp.Symbols('SOME_DB')
        >>> symbols['PARAM'] = 5
        >>> ticks = otp.DataSource('SOME_DB', tick_type='TT')
        >>> ticks['SYMBOL_PARAM'] = ticks.Symbol['PARAM', int] + 1
        >>> ticks['SYMBOL_PARAM'].dtype
        <class 'int'>
        >>> ticks = otp.merge([ticks], symbols=symbols)
        >>> otp.run(ticks)
                             Time  X  SYMBOL_PARAM
        0 2003-12-01 00:00:00.000  1             6
        1 2003-12-01 00:00:00.000 -3             6
        2 2003-12-01 00:00:00.001  2             6
        3 2003-12-01 00:00:00.001 -2             6
        4 2003-12-01 00:00:00.002  3             6
        5 2003-12-01 00:00:00.002 -1             6

        It also works with :py:class:`~onetick.py.msectime` and :py:class:`~onetick.py.nsectime` types:

        >>> symbols = otp.Symbols('SOME_DB')
        >>> symbols['NSECTIME_PARAM'] = symbols['Time'] + otp.Nano(100)
        >>> symbols['MSECTIME_PARAM'] = symbols['Time'] + otp.Milli(1)
        >>> ticks = otp.DataSource('SOME_DB', tick_type='TT')
        >>> ticks['NSECTIME_PARAM'] = ticks.Symbol['NSECTIME_PARAM', otp.nsectime] + otp.Nano(1)
        >>> ticks['MSECTIME_PARAM'] = ticks.Symbol['MSECTIME_PARAM', otp.msectime] + otp.Milli(1)
        >>> ticks['NSECTIME_PARAM'].dtype
        <class 'onetick.py.types.nsectime'>
        >>> ticks['MSECTIME_PARAM'].dtype
        <class 'onetick.py.types.msectime'>
        >>> ticks = otp.merge([ticks], symbols=symbols)
        >>> otp.run(ticks)
                             Time  X                NSECTIME_PARAM          MSECTIME_PARAM
        0 2003-12-01 00:00:00.000  1 2003-12-01 00:00:00.000000101 2003-12-01 00:00:00.002
        1 2003-12-01 00:00:00.000 -3 2003-12-01 00:00:00.000000101 2003-12-01 00:00:00.002
        2 2003-12-01 00:00:00.001  2 2003-12-01 00:00:00.000000101 2003-12-01 00:00:00.002
        3 2003-12-01 00:00:00.001 -2 2003-12-01 00:00:00.000000101 2003-12-01 00:00:00.002
        4 2003-12-01 00:00:00.002  3 2003-12-01 00:00:00.000000101 2003-12-01 00:00:00.002
        5 2003-12-01 00:00:00.002 -1 2003-12-01 00:00:00.000000101 2003-12-01 00:00:00.002
        """
        if not isinstance(item, tuple):
            raise ValueError(f"tuple should be passed, but {type(item)} was passed")
        if len(item) != 2:
            raise ValueError(f"tuple's length should be 2 but it is {len(item)}")
        item, dtype = item
        return self._get(item, dtype)


Symbol = SymbolType()  # noqa mypy fix



import re
from collections import OrderedDict
from abc import ABC, abstractmethod
from typing import List, Type


def get_properties(declaration):
    """
    Construct dictionary from properties in format
    <prop_1=value_1 prop_2=value_2 ... >
    or
    <prop_1=value_1 prop_2=value_2 ... />
    and return it back
    """
    tokens = declaration.split()

    properties = {
        token.split("=")[0].lower(): token.split("=")[1].replace("/>", "").replace(">", "")
        for token in tokens[1:]
        if "=" in token
    }

    return properties


class Generator:

    def __iter__(self):
        return self

    def __next__(self):
        self.current = next(self.generator)
        return self.current


class Action(ABC):

    def __init__(self):
        self.conditions = OrderedDict()
        # additional conditions that we can deduce
        # it's applicable for the sections without properties
        #
        # For example, if you add DB, then you can forget that
        # it's situated in the single <databases> section; that
        # information we retrieve from the entities inheritance
        # put here and use along with conditions dict
        self._add_conditions = OrderedDict()

        self.executed = False
        self.just_executed = False

    @abstractmethod
    def do(self):
        pass

    def clone(self, *args, **kwargs):
        obj = self.__class__(*args, **kwargs)
        obj.conditions = dict(self.conditions)

        for key, _ in obj.conditions.items():
            obj.conditions[key] = False

        return obj

    def apply(self, tag, entity):
        if self.executed:
            self.just_executed = False
            return

        # first time it updates and then does nothing
        self.conditions.update(self._add_conditions)
        self._add_conditions = OrderedDict()

        tag = tag.lower()

        for key, value in self.conditions.items():
            t, f, v = key

            if tag == t:
                if f:
                    if hasattr(entity, f) and getattr(entity, f) == '"' + v + '"':
                        self.conditions[(t, f, v)] = True
                    else:
                        # if we move to the same similar section, but with another key,
                        # then set it to False, because in the previous section it might be
                        # set to True
                        self.conditions[(t, f, v)] = False
                else:
                    self.conditions[(t, f, v)] = True
            else:
                if value:
                    continue

            if not self.conditions[(t, f, v)]:
                break

        self.executed = self.is_result()
        self.just_executed = self.executed

    def is_result(self):
        for v in self.conditions.values():
            if not v:
                return False

        return True

    def add_where(self, tag, **kwargs):
        if not isinstance(tag, str):
            # for case when it's class name
            tag = tag.TAG

        if len(kwargs) > 1:
            raise ValueError("It is not support to set multiple keys yet!")

        if len(kwargs) == 0:
            self.conditions[(tag.lower(), None, None)] = False
        else:
            self.conditions[(tag.lower(), list(kwargs.keys())[0].lower(), list(kwargs.values())[0])] = False


def apply_actions(func, reader, writer, actions=None, flush=False):
    from locator_parser.io import LinesReader
    from locator_parser.actions import DoNothing

    if actions is None:
        actions = []

    if len(actions) == 0:
        actions.append(DoNothing())

    func(reader, writer, actions[0])

    for action in actions[1:]:
        lines = "\n".join(map(lambda x: x.replace("\n", ""), writer.lines))
        new_reader = LinesReader(lines)

        func(new_reader, writer, action)

    if flush:
        writer.flush()

    for action in actions:
        if not action.executed:
            return False

    return True


class Writer(ABC):

    NOTHING_MSG = 0
    NEW_MSG = 1
    MODIFY_MSG = 2
    DELETE_MSG = 3

    def __init__(self):
        self.refresh()

    def refresh(self):
        self.__gen = self.__generator__()
        # init generator to send further messages
        next(self.__gen)

        self.lines = []

    def __generator__(self):
        while True:
            msg_type = yield

            if msg_type == Writer.NOTHING_MSG:
                # since second value is a just next line
                line = yield
                self.lines.append(line)
            elif msg_type == Writer.NEW_MSG:
                lines = yield

                assert isinstance(lines, list)

                self.lines = self.lines[:-1] + lines + self.lines[-1:]
            elif msg_type == Writer.MODIFY_MSG:
                pos, orig_lines, new_lines = yield

                for inx, pair in enumerate(zip(orig_lines, new_lines)):
                    orig_line, new_line = pair

                    self.lines[-pos + inx] = self.lines[-pos + inx].replace(orig_line, new_line)
            elif msg_type == Writer.DELETE_MSG:
                lines_to_delete = yield

                for _ in range(len(lines_to_delete)):
                    self.lines.pop()

    def __next__(self):
        return self.__send(None)

    def __send(self, arg):
        """private method to not allow user to send something directly"""
        self.__gen.send(arg)

    def put_as_is(self, line):
        self.__send(Writer.NOTHING_MSG)
        self.__send(line)

    def add(self, lines):
        self.__send(Writer.NEW_MSG)
        self.__send(lines)

    def modify(self, pos, orig_lines, new_lines):
        self.__send(Writer.MODIFY_MSG)
        self.__send((pos, orig_lines, new_lines))

    def delete(self, lines):
        self.__send(Writer.DELETE_MSG)
        self.__send(lines)

    @abstractmethod
    def flush(self):
        pass


class Reader(Generator):

    def __init__(self):
        self.generator = self.__generator__()
        self.writer = None
        self.current = None

    def set_writer(self, writer):
        self.__writer = writer

    def __generator__(self):
        for line in self.iterable_object:
            # put just coming line as is into the writer
            self.__writer.put_as_is(line)

            line = line.strip()

            if line == "" or line[0:1] == "#":
                continue

            yield line


class Entity(Generator, ABC):

    # Tag to identify entity
    TAG = ""
    # Whether entity has properties
    HAS_PROPERTIES = False
    # Whether it is single line properties,
    # it means that entity ends with '/>' instead of '</TAG>'
    SINGLE = False
    # children entities
    CHILDREN: List[Type["Entity"]] = []

    def __init__(self, **kwargs):
        if self.__class__.SINGLE:
            self.__class__.HAS_PROPERTIES = True

        self.__tag = self.__class__.TAG.lower()

        # define condition to identify section
        pattern = "<\\s*" + self.__tag
        if self.__class__.HAS_PROPERTIES:
            pattern += "\\s+"
        else:
            pattern += "\\s*>"

        self.__cond = lambda x: re.match(pattern, x.lower())
        self.__lines = []
        self.__properties_lines = []
        self.__do = False
        self.__custom_properties = kwargs

    def to_lines(self):
        result = []
        line = "<" + self.__class__.TAG + " "

        for key, value in self.__custom_properties.items():
            line += key + '="' + str(value) + '" '

        if self.__class__.SINGLE:
            line += "/>"
            result.append(line)
        else:
            line += ">"
            result.append(line)

            offset = "    "
            for child in self.__class__.CHILDREN:
                if not child.SINGLE and not child.HAS_PROPERTIES:
                    result.append(offset + "<" + child.TAG + ">")
                    result.append(offset + "</" + child.TAG + ">")

            result.append("</" + self.__class__.TAG + ">")

        return result

    def __process(self, reader, writer, action):
        line = reader.current
        pattern = "</\\s*" + self.__tag + "\\s*>"

        while True:
            self.__lines.append(line)

            if self.__class__.SINGLE:
                if line.find("/>") != -1:
                    # the final line for single and we have to return it
                    yield line

                    if action.just_executed:
                        self.finish(writer, action)

                    return
            else:
                if re.match(pattern, line.lower()):
                    self.finish(writer, action)
                    return

            try:
                line = next(reader)
                yield line
            except StopIteration:
                return

    def finish(self, writer, action):
        if self.__do:
            action.do(self, self.__lines, self.__properties_lines, writer)

    def __read_properties__(self):
        line = self.current
        self.__properties_lines.append(line)
        declare = line

        while line.find(">") == -1:
            line = next(self)
            self.__properties_lines.append(line)
            declare += " " + line

        properties = get_properties(declare)
        self.__dict__.update(properties)

    def __process_children__(self, reader, writer, action):
        for _ in self:
            for child_p in self.__class__.CHILDREN:
                child_p()(self, writer, action)

    def __call__(self, reader, writer, action):
        if self.__cond(reader.current):
            self.generator = self.__process(reader, writer, action)
            self.current = reader.current

            self.__read_properties__()
            action.apply(self.__class__.TAG, self)

            if action.just_executed:
                self.__do = True

            self.parse(self, writer, action)

            self.__process_children__(self, writer, action)

            return True

        return False

    def parse(self, reader, writer, action):
        pass



from typing import Union

from onetick.py import configuration, utils
from onetick.py.core.column_operations.accessors._accessor import _Accessor
from onetick.py.backports import Literal
from onetick.py.types import datetime, value2str
from onetick.py.docs.utils import docstring, param_doc
from onetick.py.core.column import _Column
from onetick.py.core.column_operations.base import _Operation


_timezone_doc = param_doc(
    name='timezone',
    str_annotation='str | Operation | Column',
    desc="""
    Name of the timezone, an operation or a column with it.
    By default, the timezone of the query will be used.
    """,
    annotation=Union[str, _Operation, _Column]
)


class _DtAccessor(_Accessor):

    """
    Accessor for datetime functions

    >>> data = otp.Ticks(X=[otp.dt(2019, 1, 1, 1, 1, 1), otp.dt(2019, 2, 2, 2, 2, 2)])
    >>> data["Y"] = data["X"].dt.<function_name>()  # doctest: +SKIP
    """

    @docstring(parameters=[_timezone_doc], add_self=True)
    def strftime(self, format='%Y/%m/%d %H:%M:%S.%J', timezone=None):
        """
        Converts the number of nanoseconds (datetime) since 1970/01/01 GMT into
        the string specified by ``format`` for a specified ``timezone``.

        Parameters
        ----------
        format: str
            The format might contain any characters, but the following combinations of
            characters have special meanings

            %Y - Year (4 digits)

            %y - Year (2 digits)

            %m - Month (2 digits)

            %d - Day of month (2 digits)

            %H - Hours (2 digits, 24-hour format)

            %I - Hours (2 digits, 12-hour format)

            %M - Minutes (2 digits)

            %S - Seconds (2 digits)

            %q - Milliseconds (3 digits)

            %J - Nanoseconds (9 digits)

            %p - AM/PM (2 characters)

            %% - % character

        Examples
        --------
        >>> t = otp.Ticks(A=[otp.dt(2019, 1, 1, 1, 1, 1), otp.dt(2019, 2, 2, 2, 2, 2)])
        >>> t['B'] = t['A'].dt.strftime('%d.%m.%Y')
        >>> otp.run(t)[['A', 'B']]
                            A           B
        0 2019-01-01 01:01:01  01.01.2019
        1 2019-02-02 02:02:02  02.02.2019
        """
        if timezone is utils.default:
            timezone = configuration.config.tz
        timezone, format_str = self._preprocess_tz_and_format(timezone, format)

        def formatter(x):
            return f'nsectime_format({format_str},{x},{timezone})'

        return _DtAccessor.Formatter(self._base_column,
                                     str,
                                     formatter=formatter)

    def date(self):
        """
        Return a new :py:class:`onetick.py.nsectime` type operation filled with date only.

        Examples
        --------
        >>> data = otp.Ticks(X=[otp.dt(2019, 1, 1, 1, 1, 1), otp.dt(2019, 2, 2, 2, 2, 2)])
        >>> data["X"] = data["X"].dt.date()     # OTdirective: snippet-name: timestamp operations.date;
        >>> df = otp.run(data, timezone="GMT")
        >>> df["X"]
        0   2019-01-01
        1   2019-02-02
        Name: X, dtype: datetime64[ns]
        """
        format_str = "%Y%m%d"
        return self.strftime(format_str, None).str.to_datetime(format_str, None)

    @docstring(parameters=[_timezone_doc], add_self=True)
    def day_of_week(self, start_index: int = 1, start_day: Literal['monday', 'sunday'] = 'monday', timezone=None):
        """
        Return the day of the week.

        Assuming the week starts on ``start_day``, which is denoted by ``start_index``.
        Default: Monday - 1, ..., Sunday - 7; set according to ISO8601

        Parameters
        ----------
        start_index: int
            Sunday index.
        start_day: 'monday' or 'sunday'
            Day that will be denoted with ``start_index``

        Examples
        --------
        >>> data = otp.Ticks(X=[otp.dt(2022, 5, i) for i in range(10, 17)])
        >>> data['DAY_OF_WEEK'] = data['X'].dt.day_of_week()
        >>> otp.run(data)[['X', 'DAY_OF_WEEK']]
                   X  DAY_OF_WEEK
        0 2022-05-10            2
        1 2022-05-11            3
        2 2022-05-12            4
        3 2022-05-13            5
        4 2022-05-14            6
        5 2022-05-15            7
        6 2022-05-16            1
        """

        if start_day not in ['monday', 'sunday']:
            raise ValueError(f"'start_day' parameter ({start_day}) not in ['monday', 'sunday']")

        timezone, _ = self._preprocess_tz_and_format(timezone, '')

        def formatter(x):
            format_ = f'day_of_week({x},{timezone})'
            if start_day == 'monday':
                # CASE should be uppercased because it can be used in per-tick script
                format_ = f'CASE({format_}, 0, 7, {format_})-1'
            format_ += f'+{start_index}'
            return format_

        return _DtAccessor.Formatter(self._base_column,
                                     int,
                                     formatter=formatter)

    @docstring(parameters=[_timezone_doc], add_self=True)
    def day_name(self, timezone=None):
        """
        Returns the name of the weekday.

        Examples
        --------
        >>> data = otp.Ticks(X=[otp.dt(2022, 5, i) for i in range(10, 17)])
        >>> data['DAY_NAME'] = data['X'].dt.day_name()
        >>> otp.run(data)[['X', 'DAY_NAME']]
                   X   DAY_NAME
        0 2022-05-10    Tuesday
        1 2022-05-11  Wednesday
        2 2022-05-12   Thursday
        3 2022-05-13     Friday
        4 2022-05-14   Saturday
        5 2022-05-15     Sunday
        6 2022-05-16     Monday
        """
        timezone, _ = self._preprocess_tz_and_format(timezone, '')

        return _DtAccessor.Formatter(self._base_column,
                                     str,
                                     formatter=lambda x: f'DAYNAME({x},{timezone})')

    @docstring(parameters=[_timezone_doc], add_self=True)
    def day_of_month(self, timezone=None):
        """
        Return the day of the month.

        Examples
        --------
        >>> data = otp.Ticks(X=[otp.dt(2022, 5, i) for i in range(10, 17)])
        >>> data['DAY_OF_MONTH'] = data['X'].dt.day_of_month()
        >>> otp.run(data)[['X', 'DAY_OF_MONTH']]
                   X  DAY_OF_MONTH
        0 2022-05-10            10
        1 2022-05-11            11
        2 2022-05-12            12
        3 2022-05-13            13
        4 2022-05-14            14
        5 2022-05-15            15
        6 2022-05-16            16
        """
        timezone, _ = self._preprocess_tz_and_format(timezone, '')

        return _DtAccessor.Formatter(self._base_column,
                                     int,
                                     formatter=lambda x: f'DAYOFMONTH({x},{timezone})')

    @docstring(parameters=[_timezone_doc], add_self=True)
    def day_of_year(self, timezone=None):
        """
        Return the day of the year.

        Examples
        --------
        >>> data = otp.Ticks(X=[otp.dt(2022, 5, i) for i in range(10, 17)])
        >>> data['DAY_OF_YEAR'] = data['X'].dt.day_of_year()
        >>> otp.run(data)[['X', 'DAY_OF_YEAR']]
                   X  DAY_OF_YEAR
        0 2022-05-10          130
        1 2022-05-11          131
        2 2022-05-12          132
        3 2022-05-13          133
        4 2022-05-14          134
        5 2022-05-15          135
        6 2022-05-16          136
        """
        timezone, _ = self._preprocess_tz_and_format(timezone, '')

        return _DtAccessor.Formatter(self._base_column,
                                     int,
                                     formatter=lambda x: f'DAYOFYEAR({x},{timezone})')

    @docstring(parameters=[_timezone_doc], add_self=True)
    def hour(self, timezone=None):
        """
        Return the hour.

        Examples
        --------
        >>> data = otp.Ticks(X=[otp.dt(2022, 5, 1, i, 0, 6) for i in range(10, 17)])
        >>> data['HOUR'] = data['X'].dt.hour()
        >>> otp.run(data)[['X', 'HOUR']]
                            X  HOUR
        0 2022-05-01 10:00:06    10
        1 2022-05-01 11:00:06    11
        2 2022-05-01 12:00:06    12
        3 2022-05-01 13:00:06    13
        4 2022-05-01 14:00:06    14
        5 2022-05-01 15:00:06    15
        6 2022-05-01 16:00:06    16
        """
        timezone, _ = self._preprocess_tz_and_format(timezone, '')

        return _DtAccessor.Formatter(self._base_column,
                                     int,
                                     formatter=lambda x: f'HOUR({x},{timezone})')

    @docstring(parameters=[_timezone_doc], add_self=True)
    def minute(self, timezone=None):
        """
        Return the minute.


        Examples
        --------
        >>> data = otp.Ticks(X=[otp.dt(2022, 5, 1, 15, i, 6) for i in range(10, 17)])
        >>> data['MINUTE'] = data['X'].dt.minute()
        >>> otp.run(data)[['X', 'MINUTE']]
                            X  MINUTE
        0 2022-05-01 15:10:06      10
        1 2022-05-01 15:11:06      11
        2 2022-05-01 15:12:06      12
        3 2022-05-01 15:13:06      13
        4 2022-05-01 15:14:06      14
        5 2022-05-01 15:15:06      15
        6 2022-05-01 15:16:06      16
        """
        timezone, _ = self._preprocess_tz_and_format(timezone, '')

        return _DtAccessor.Formatter(self._base_column,
                                     int,
                                     formatter=lambda x: f'MINUTE({x},{timezone})')

    @docstring(parameters=[_timezone_doc], add_self=True)
    def second(self, timezone=None):
        """
        Return the second.

        Examples
        --------
        >>> data = otp.Ticks(X=[otp.dt(2022, 5, 1, 15, 11, i) for i in range(10, 17)])
        >>> data['SECOND'] = data['X'].dt.second()
        >>> otp.run(data)[['X', 'SECOND']]
                            X  SECOND
        0 2022-05-01 15:11:10      10
        1 2022-05-01 15:11:11      11
        2 2022-05-01 15:11:12      12
        3 2022-05-01 15:11:13      13
        4 2022-05-01 15:11:14      14
        5 2022-05-01 15:11:15      15
        6 2022-05-01 15:11:16      16
        """
        timezone, _ = self._preprocess_tz_and_format(timezone, '')

        return _DtAccessor.Formatter(self._base_column,
                                     int,
                                     formatter=lambda x: f'SECOND({x},{timezone})')

    @docstring(parameters=[_timezone_doc], add_self=True)
    def month(self, timezone=None):
        """
        Return the month.

        Examples
        --------
        >>> data = otp.Ticks(X=[otp.dt(2022, i, 1) for i in range(3, 11)])
        >>> data['MONTH'] = data['X'].dt.month()
        >>> otp.run(data)[['X', 'MONTH']]
                   X  MONTH
        0 2022-03-01      3
        1 2022-04-01      4
        2 2022-05-01      5
        3 2022-06-01      6
        4 2022-07-01      7
        5 2022-08-01      8
        6 2022-09-01      9
        7 2022-10-01     10
        """
        timezone, _ = self._preprocess_tz_and_format(timezone, '')

        return _DtAccessor.Formatter(self._base_column,
                                     int,
                                     formatter=lambda x: f'MONTH({x},{timezone})')

    @docstring(parameters=[_timezone_doc], add_self=True)
    def month_name(self, timezone=None):
        """
        Return name of the month.

        Examples
        --------
        >>> data = otp.Ticks(X=[otp.dt(2022, i, 1) for i in range(3, 11)])
        >>> data['MONTH_NAME'] = data['X'].dt.month_name()
        >>> otp.run(data)[['X', 'MONTH_NAME']]
                   X MONTH_NAME
        0 2022-03-01        Mar
        1 2022-04-01        Apr
        2 2022-05-01        May
        3 2022-06-01        Jun
        4 2022-07-01        Jul
        5 2022-08-01        Aug
        6 2022-09-01        Sep
        7 2022-10-01        Oct
        """
        timezone, _ = self._preprocess_tz_and_format(timezone, '')

        return _DtAccessor.Formatter(self._base_column,
                                     str,
                                     formatter=lambda x: f'MONTHNAME({x},{timezone})')

    @docstring(parameters=[_timezone_doc], add_self=True)
    def quarter(self, timezone=None):
        """
        Return the quarter.

        Examples
        --------
        >>> data = otp.Ticks(X=[otp.dt(2022, i, 1) for i in range(3, 11)])
        >>> data['QUARTER'] = data['X'].dt.quarter()
        >>> otp.run(data)[['X', 'QUARTER']]
                   X  QUARTER
        0 2022-03-01        1
        1 2022-04-01        2
        2 2022-05-01        2
        3 2022-06-01        2
        4 2022-07-01        3
        5 2022-08-01        3
        6 2022-09-01        3
        7 2022-10-01        4
        """
        timezone, _ = self._preprocess_tz_and_format(timezone, '')

        return _DtAccessor.Formatter(self._base_column,
                                     int,
                                     formatter=lambda x: f'QUARTER({x},{timezone})')

    @docstring(parameters=[_timezone_doc], add_self=True)
    def year(self, timezone=None):
        """
        Return the year.

        Examples
        --------
        >>> data = otp.Ticks(X=[otp.dt(2020 + i, 3, 1) for i in range(3, 11)])
        >>> data['YEAR'] = data['X'].dt.year()
        >>> otp.run(data)[['X', 'YEAR']]
                   X  YEAR
        0 2023-03-01  2023
        1 2024-03-01  2024
        2 2025-03-01  2025
        3 2026-03-01  2026
        4 2027-03-01  2027
        5 2028-03-01  2028
        6 2029-03-01  2029
        7 2030-03-01  2030
        """
        timezone, _ = self._preprocess_tz_and_format(timezone, '')

        return _DtAccessor.Formatter(self._base_column,
                                     int,
                                     formatter=lambda x: f'YEAR({x},{timezone})')

    @docstring(parameters=[_timezone_doc], add_self=True)
    def date_trunc(self,
                   date_part: Literal['year', 'quarter', 'month', 'week', 'day', 'hour', 'minute', 'second',
                                      'millisecond', 'nanosecond'],
                   timezone=None):
        """
        Truncates to the specified precision.

        Parameters
        ----------
        date_part: str | Operation | Column
            Precision to truncate datetime to. Possible values are 'year', 'quarter', 'month', 'week', 'day', 'hour',
            'minute', 'second', 'millisecond' and 'nanosecond'.
            Notice that beginning of week is considered to be Sunday.

        Examples
        --------
        >>> data = otp.Ticks(X=[otp.dt(2020, 11, 11, 5, 4, 13, 101737, 879)] * 7,
        ...                  DATE_PART=['year', 'day', 'hour', 'minute', 'second', 'millisecond', 'nanosecond'])
        >>> data['TRUNCATED_X'] = data['X'].dt.date_trunc(data['DATE_PART'])
        >>> otp.run(data)[['X', 'TRUNCATED_X', 'DATE_PART']]
                                      X                   TRUNCATED_X    DATE_PART
        0 2020-11-11 05:04:13.101737879 2020-01-01 00:00:00.000000000         year
        1 2020-11-11 05:04:13.101737879 2020-11-11 00:00:00.000000000          day
        2 2020-11-11 05:04:13.101737879 2020-11-11 05:00:00.000000000         hour
        3 2020-11-11 05:04:13.101737879 2020-11-11 05:04:00.000000000       minute
        4 2020-11-11 05:04:13.101737879 2020-11-11 05:04:13.000000000       second
        5 2020-11-11 05:04:13.101737879 2020-11-11 05:04:13.101000000  millisecond
        6 2020-11-11 05:04:13.101737879 2020-11-11 05:04:13.101737879   nanosecond
        """
        timezone, _ = self._preprocess_tz_and_format(timezone, '')
        date_part = value2str(date_part)

        return _DtAccessor.Formatter(self._base_column,
                                     datetime,
                                     formatter=lambda x: f'DATE_TRUNC({date_part},{x},{timezone})')

    @docstring(parameters=[_timezone_doc], add_self=True)
    def week(self, timezone=None):
        """
        Returns the week.

        Examples
        --------
        >>> data = otp.Ticks(X=[otp.dt(2020, i, 1) for i in range(3, 11)])
        >>> data['WEEK'] = data['X'].dt.week()
        >>> otp.run(data)[['X', 'WEEK']]
                   X  WEEK
        0 2020-03-01    10
        1 2020-04-01    14
        2 2020-05-01    18
        3 2020-06-01    23
        4 2020-07-01    27
        5 2020-08-01    31
        6 2020-09-01    36
        7 2020-10-01    40
        """
        timezone, _ = self._preprocess_tz_and_format(timezone, '')

        return _DtAccessor.Formatter(self._base_column,
                                     int,
                                     formatter=lambda x: f'WEEK({x},{timezone})')



import ast
import inspect
import textwrap
import types
import operator
import tokenize
import warnings
from typing import Callable, Union, Any, Optional, Iterable, Type, Tuple, Dict, List, TypeVar
from copy import deepcopy
from functools import wraps

from onetick.py.backports import astunparse, cached_property

from collections import deque
from contextlib import contextmanager

from .. import types as ott
from .column_operations.base import _Operation
from .column import _Column
from .lambda_object import _EmulateObject, _LambdaIfElse, _EmulateStateVars, _CompareTrackScope
from ._internal._state_objects import (
    _TickSequence, _TickSequenceTickBase, _TickListTick, _TickSetTick, _TickDequeTick, _DynamicTick
)


class Static:
    """
    Class for declaring static local variable in per-tick script.
    Static variables are defined once and save their values between
    arrival of the input ticks.
    """
    def __init__(self, value):
        self.value = value

    # these functions needed mostly for linters
    def __getattr__(self, item):
        return self.value.__getattr__(item)

    def __getitem__(self, item):
        return operator.getitem(self.value, item)

    def __setitem__(self, key, value):
        return operator.setitem(self.value, key, value)


class LocalVariable(_Operation):
    """
    Class for inner representation of local variable in per-tick script.
    Only simple values are supported, tick sequences are represented by another class.
    """
    def __init__(self, name, dtype=None):
        super().__init__(op_str=f'LOCAL::{name}', dtype=dtype)
        self.name = name


class TickDescriptorFields(_TickSequence):
    """
    Class for declaring tick descriptor fields in per-tick script.
    Can only be iterated, doesn't have methods and parameters.

    See also
    --------
    :py:class:`TickDescriptorField <onetick.py.core.per_tick_script.TickDescriptorField>`

    Examples
    --------
    >>> t = otp.Tick(A=1)
    >>> def fun(tick):
    ...     for field in otp.tick_descriptor_fields():
    ...         tick['NAME'] = field.get_name()
    >>> t = t.script(fun)
    >>> otp.run(t)
            Time  A NAME
    0 2003-12-01  1    A
    """
    def __init__(self):
        # we don't want to inherit parent class init
        pass

    def __str__(self):
        return 'LOCAL::INPUT_TICK_DESCRIPTOR_FIELDS'

    @property
    def _tick_class(self):
        return TickDescriptorField


def tick_list_tick():
    """
    Can be used only in per-tick script function
    to define a tick list tick local variable.

    Tick list ticks can be used with some methods
    of tick lists :py:class:`onetick.py.state.tick_list`.

    See also
    --------
    :py:class:`onetick.py.state.tick_list`.

    Returns
    -------
    :py:class:`onetick.py.static` value with tick object.

    Examples
    --------
    >>> def fun(tick):
    ...    t = otp.tick_list_tick()
    ...    tick.state_vars['LIST'].push_back(t)
    """
    return Static(_TickListTick(None))


def tick_set_tick():
    """
    Can be used only in per-tick script function
    to define a tick set tick local variable.

    Tick set ticks can be used with some methods
    of tick sets :py:class:`onetick.py.state.tick_set`.

    See also
    --------
    :py:class:`onetick.py.state.tick_set`.

    Returns
    -------
    :py:class:`onetick.py.static` value with tick object.

    Examples
    --------
    >>> def fun(tick):
    ...    t = otp.tick_set_tick()
    ...    if tick.state_vars['SET'].find(t, -1):
    ...        tick['RES'] = '-1'
    """
    return Static(_TickSetTick(None))


def tick_deque_tick():
    """
    Can be used only in per-tick script function
    to define a tick deque tick local variable.

    Tick deque ticks can be used with some methods
    of tick deques :py:class:`onetick.py.state.tick_deque`.

    See also
    --------
    :py:class:`onetick.py.state.tick_deque`.

    Returns
    -------
    :py:class:`onetick.py.static` value with tick object.

    Examples
    --------
    >>> def fun(tick):
    ...    t = otp.tick_deque_tick()
    ...    tick.state_vars['DEQUE'].get_tick(0, t)
    """
    return Static(_TickDequeTick(None))


def dynamic_tick():
    """
    Can be used only in per-tick script function
    to define a dynamic tick local variable.

    Dynamic ticks can be used with some methods
    of all tick sequences.

    See also
    --------
    :py:class:`onetick.py.state.tick_list`
    :py:class:`onetick.py.state.tick_set`
    :py:class:`onetick.py.state.tick_deque`

    Returns
    -------
    :py:class:`onetick.py.static` value with tick object.

    Examples
    --------
    >>> def fun(tick):
    ...    t = otp.dynamic_tick()
    ...    t['X'] = tick['SUM']
    """
    return Static(_DynamicTick(None))


class TickDescriptorField(_TickSequenceTickBase):
    """
    Tick descriptor field object.
    Can be accessed only while iterating over
    :py:class:`otp.tick_descriptor_fields <onetick.py.core.per_tick_script.TickDescriptorFields>`
    in per-tick script.

    Examples
    --------
    >>> t = otp.Tick(A=2, B='B', C=1.2345)
    >>> def fun(tick):
    ...     tick['NAMES'] = ''
    ...     tick['TYPES'] = ''
    ...     tick['SIZES'] = ''
    ...     for field in otp.tick_descriptor_fields():
    ...         tick['NAMES'] += field.get_name() + ','
    ...         tick['TYPES'] += field.get_type() + ','
    ...         tick['SIZES'] += field.get_size().apply(str) + ','
    >>> t = t.script(fun)
    >>> otp.run(t)
            Time  A  B       C   NAMES                TYPES    SIZES
    0 2003-12-01  2  B  1.2345  A,B,C,  long,string,double,  8,64,8,
    """

    _definition = 'TICK_DESCRIPTOR_FIELD'

    def get_field_name(self):
        """
        Get the name of the field.

        Returns
        -------
        onetick.py.Operation
        """
        return _Operation(op_str=f'{self}.GET_FIELD_NAME()', dtype=str)

    def get_name(self):
        """
        Get the name of the field.

        Returns
        -------
        onetick.py.Operation
        """
        return self.get_field_name()

    def get_size(self):
        """
        Get the size of the type of the field.

        Returns
        -------
        onetick.py.Operation
        """
        return _Operation(op_str=f'{self}.GET_SIZE()', dtype=int)

    def get_type(self):
        """
        Get the name of the type of the field.

        Returns
        -------
        onetick.py.Operation
        """
        return _Operation(op_str=f'{self}.GET_TYPE()', dtype=str)


class Expression:
    """
    Class to save per-tick-script expressions along with their possible values.

    Parameters
    ----------
    expr
        string expression that will be saved to per tick script
    values:
        values that this expression can take.
        For example, bool operation can take many values.
    lhs:
        True if expression is left hand expression.
        In this case value of expression must be callable.
        Calling it with right hand expression value as an argument
        should be the same as execute the whole expression.
    """
    def __init__(self, *values: Any, expr: Optional[str] = None, lhs: bool = False):
        self.values = values
        self._expr = expr
        self.lhs = lhs
        if self.lhs:
            assert callable(self.value)
            assert self.expr

    @property
    def expr(self):
        if self._expr:
            return self._expr
        if self.is_emulator:
            self._expr = 'LOCAL::OUTPUT_TICK'
        elif self.is_column:
            self._expr = str(self.value)
        elif self.values:
            self._expr = self.value_to_onetick(self.value)
        return self._expr

    @property
    def value(self):
        length = len(self.values)
        if length == 0:
            raise ValueError(f"Expression '{self}' doesn't have values.")
        if length > 1:
            raise ValueError(f"Expression '{self}' have more than one value.")
        return self.values[0]

    @cached_property
    def dtype(self):
        return ott.get_type_by_objects(self.values)

    @property
    def is_emulator(self) -> bool:
        try:
            return isinstance(self.value, _EmulateObject)
        except ValueError:
            return False

    @property
    def is_state_vars(self) -> bool:
        try:
            return isinstance(self.value, _EmulateStateVars)
        except ValueError:
            return False

    @property
    def is_static(self) -> bool:
        try:
            return isinstance(self.value, Static)
        except ValueError:
            return False

    @property
    def is_dynamic_tick(self) -> bool:
        try:
            return isinstance(self.value, _DynamicTick)
        except ValueError:
            return False

    @property
    def is_tick(self) -> bool:
        try:
            return isinstance(self.value, _TickSequenceTickBase)
        except ValueError:
            return False

    @property
    def is_column(self) -> bool:
        try:
            return isinstance(self.value, _Column)
        except ValueError:
            return False

    @property
    def is_local_variable(self) -> bool:
        try:
            return isinstance(self.value, LocalVariable)
        except ValueError:
            return False

    @property
    def is_operation(self) -> bool:
        try:
            return isinstance(self.value, _Operation)
        except ValueError:
            return False

    @property
    def predefined(self) -> bool:
        """Check if the value of expression is known before the execution of query"""
        return not self.is_operation

    @property
    def expressible(self) -> bool:
        return bool(self.expr)

    def __str__(self):
        if not self.expressible:
            raise ValueError("This Expression can't be expressed in OneTick or is undefined yet")
        return self.expr

    def convert_to_operation(self):
        """
        Convert otp.Column to otp.Operation.
        Needed to convert expressions like:
            if tick['X']:
        to
            if (X != 0) {
        """
        if self.is_column:
            self.values = [self.value._make_python_way_bool_expression()]
            self._expr = str(self.value)

    @staticmethod
    def value_to_onetick(value: Union[str, int, float, bool, None, _Operation]) -> str:
        """
        Python value will be converted accordingly to OneTick syntax
        (lowercase boolean values, string in quotes, etc.)
        """
        if value is None:
            return str(ott.nan)
        if isinstance(value, bool):
            return str(value).lower()
        return ott.value2str(value)


class CaseOperatorParser:
    """
    Class with methods to convert ast operators to their string or python representations.
    Only ast operators that can be used in OneTick's CASE function are accepted.
    """

    @staticmethod
    def py_operator(op: Union[ast.operator, ast.cmpop, ast.unaryop, ast.boolop]) -> Callable:
        """
        Convert ast operator to python function for this operator.

        Parameters
        ----------
        op
            ast operator object
        """
        return {
            # binary
            ast.Add: operator.add,
            ast.Sub: operator.sub,
            ast.Mult: operator.mul,
            ast.Div: operator.truediv,
            ast.BitAnd: operator.and_,
            ast.BitOr: operator.or_,
            ast.Mod: operator.mod,
            # unary
            ast.UAdd: operator.pos,
            ast.USub: operator.neg,
            ast.Not: operator.not_,
            ast.Invert: operator.invert,
            # compare
            ast.Lt: operator.lt,
            ast.LtE: operator.le,
            ast.Gt: operator.gt,
            ast.GtE: operator.ge,
            ast.Eq: operator.eq,
            ast.NotEq: operator.ne,
            # bool
            ast.And: lambda x, y: x and y,
            ast.Or: lambda x, y: x or y,
        }[type(op)]  # type: ignore[return-value]

    @staticmethod
    def operator(op: Union[ast.operator, ast.cmpop, ast.unaryop, ast.boolop]) -> str:
        """
        Convert ast operator to OneTick's string representation.

        Parameters
        ----------
        op
            ast operator object
        """
        return {
            # binary
            ast.Add: '+',
            ast.Sub: '-',
            ast.Mult: '*',
            ast.Div: '/',
            ast.Mod: '%',
            # unary
            ast.UAdd: '+',
            ast.USub: '-',
            # compare
            ast.Lt: '<',
            ast.LtE: '<=',
            ast.Gt: '>',
            ast.GtE: '>=',
            ast.Eq: '=',
            ast.NotEq: '!=',
            # bool
            ast.And: 'AND',
            ast.Or: 'OR',
        }[type(op)]


class OperatorParser(CaseOperatorParser):
    """
    Class with methods to convert ast operators to their string or python representations.
    Only ast operators that can be used in OneTick's per tick script are accepted.
    """

    @staticmethod
    def py_operator(op: Union[ast.operator, ast.cmpop, ast.unaryop, ast.boolop],
                    aug: bool = False, **kwargs) -> Callable:
        """
        Convert ast operator to python function for this operator.

        Parameters
        ----------
        op
            ast operator object
        aug
            ast don't have separate inplace operators (+=, -=, etc.)
            If this parameter is True then operator is inplace and otherwise if False.
        """
        if aug:
            assert isinstance(op, ast.operator)
            return {
                ast.Add: operator.iadd,
                ast.Sub: operator.isub,
                ast.Mult: operator.imul,
                ast.Div: operator.itruediv,
            }[type(op)]
        return CaseOperatorParser.py_operator(op, **kwargs)

    @staticmethod
    def operator(op: Union[ast.operator, ast.cmpop, ast.unaryop, ast.boolop],
                 aug: bool = False) -> str:
        """
        Convert ast operator to its string representation.

        Parameters
        ----------
        op
            ast operator object
        aug
            ast don't have separate inplace binary operators (+=, -=, etc.)
            If this parameter is True then parameter is inplace and otherwise if False.
        """
        if aug:
            assert isinstance(op, ast.operator)
            return {
                ast.Add: '+=',
                ast.Sub: '-=',
                ast.Mult: '*=',
                ast.Div: '/=',
            }[type(op)]
        try:
            return {
                ast.Eq: '==',
                ast.And: '&&',
                ast.Or: '||',
            }[type(op)]
        except KeyError:
            return CaseOperatorParser.operator(op)


class ExpressionParser:
    """
    Class with methods to convert ast expressions to OneTick's script or function syntax.
    """
    def __init__(self, fun: 'FunctionParser'):
        self.fun = fun
        self.operator_parser = OperatorParser()

    @contextmanager
    def _replace_context(self, closure_vars: inspect.ClosureVars):
        """
        Temporarily change closure variables in self.fun.
        Variables will be replaced with those from closure_vars parameter.
        """
        nonlocals, globals_, *_ = closure_vars
        assert isinstance(self.fun.closure_vars.globals, dict)
        assert isinstance(self.fun.closure_vars.nonlocals, dict)
        saved_globals = self.fun.closure_vars.globals.copy()
        saved_nonlocals = self.fun.closure_vars.nonlocals.copy()
        self.fun.closure_vars.globals.update(globals_)
        self.fun.closure_vars.nonlocals.update(nonlocals)
        yield
        self.fun.closure_vars.globals.update(saved_globals)
        self.fun.closure_vars.nonlocals.update(saved_nonlocals)

    def constant(self, expr: ast.Constant) -> Expression:
        """Some basic constant value: string, integer, float."""
        return Expression(expr.value)

    def string(self, expr: "ast.Str") -> Expression:
        """String (for backward compatibility with Python 3.7)."""
        return Expression(expr.s)

    def number(self, expr: "ast.Str") -> Expression:
        """Number (for backward compatibility with Python 3.7)."""
        return Expression(expr.n)

    def name(self, expr: ast.Name) -> Expression:
        """
        Name of the variable.
        Every variable in per-tick script function, if defined correctly,
        is considered to be local per-tick script variable.
        If variable with this name is not found it will be captured from function context.
        """
        if self.fun.arg_name and expr.id == self.fun.arg_name:
            value = self.fun.emulator if self.fun.emulator is not None else expr.id
            return Expression(value)

        if not isinstance(expr.ctx, ast.Load):
            # local variable, left-hand side
            return Expression(LocalVariable(expr.id))

        for dict_name in ('LOCAL_VARS', 'STATIC_VARS'):
            # local or static variable, right-hand side
            variables = getattr(self.fun.emulator, dict_name, {})
            if expr.id in variables:
                dtype = ott.get_type_by_objects([variables[expr.id]])
                if issubclass(dtype, _TickSequenceTickBase):
                    # ticks have schema, owner and methods, so using saved value
                    return Expression(variables[expr.id])
                return Expression(LocalVariable(expr.id, dtype))

        if not self.fun._from_args_annotations and expr.id in self.fun.args_annotations:
            # parameter of the per-tick script function
            return Expression(_Operation(op_str=expr.id, dtype=self.fun.args_annotations[expr.id]))

        # pylint: disable-next=eval-used
        value = eval(expr.id, self.fun.closure_vars.globals, self.fun.closure_vars.nonlocals)  # type: ignore[arg-type]
        return Expression(value)

    def index(self, expr: ast.Index) -> Expression:
        """Proxy object in ast.Subscript in python < 3.9"""
        return self.expression(expr.value)  # type: ignore[attr-defined]

    def slice(self, expr: ast.Slice) -> Expression:
        """
        Slice of the list.
        For example:
            a = [1, 2, 3, 4]
            a[2:4]
        Here, 2:4 is the slice.
        """
        lower = self.expression(expr.lower).value if expr.lower else None
        upper = self.expression(expr.upper).value if expr.upper else None
        step = self.expression(expr.step).value if expr.step else None
        return Expression(slice(lower, upper, step))

    def subscript(self, expr: ast.Subscript) -> Expression:
        """
        Expression like: tick['X'].
        Setting items of ticks and state variables is supported.
        Getting items supported for any captured variable.
        """
        val = self.expression(expr.value)
        item = self.expression(expr.slice)

        if isinstance(expr.ctx, ast.Load):
            v = val.value[item.value]
            return Expression(v)

        # index of per tick script function parameter or tick sequence tick is column name
        if not (val.is_emulator or val.is_tick or val.is_state_vars):
            raise ValueError(f"Setting items supported only for "
                             f"'{self.fun.arg_name}' function argument, "
                             f"tick sequences' ticks and state variables object")

        return Expression(
            lambda rhs: val.value.__setitem__(item.value, rhs),
            expr=item.value,
            lhs=True,
        )

    def attribute(self, expr: ast.Attribute) -> Expression:
        """
        Expression like: tick.X
        For now we only support setting attributes of first function parameter.
        Getting attributes supported for any captured variable.
        """
        val = self.expression(expr.value)
        attr = expr.attr

        if isinstance(expr.ctx, ast.Load):
            v = getattr(val.value, attr)
            return Expression(v)

        # attribute of per tick script function parameter or tick sequence tick is column name
        if not (val.is_emulator or val.is_tick):
            raise ValueError(f"Setting attributes supported only for "
                             f"'{self.fun.arg_name}' function argument and tick sequences' ticks")

        return Expression(
            lambda rhs: val.value.__setattr__(attr, rhs),
            expr=attr,
            lhs=True,
        )

    def bin_op(self, expr: ast.BinOp) -> Expression:
        """
        Binary operation expression: 2 + 2, tick['X'] * 2, etc.
        """
        left = self.expression(expr.left)
        py_op = self.operator_parser.py_operator(expr.op)
        right = self.expression(expr.right)
        value = py_op(left.value, right.value)
        return Expression(value)

    def unary_op(self, expr: ast.UnaryOp) -> Expression:
        """
        Unary operation expression: -1, -tick['X'], not tick['X'], ~tick['X'], etc.
        """
        py_op = self.operator_parser.py_operator(expr.op)
        operand = self.expression(expr.operand)
        if operand.is_operation:
            # special case for negative otp.Columns and otp.Operations
            if isinstance(expr.op, (ast.Not, ast.Invert)):
                operand.convert_to_operation()
            if isinstance(expr.op, ast.Not):
                py_op = self.operator_parser.py_operator(ast.Invert())
        value = py_op(operand.value)
        return Expression(value)

    def bool_op(self, expr: ast.BoolOp) -> Expression:
        """
        Bool operation expression: True and tick['X'], etc.
        Note that
            * all python values will be checked inplace and will not be written to the script
            * short-circuit logic will work for python values
        For example:
            True   and  0     and tick['X'] == 1  -------> false
            'true' or   False or  tick['X'] == 1  -------> true
            True   and  True  and tick['X'] == 1  -------> X == 1
        """
        value = None
        for e in expr.values:
            expression = self.expression(e)
            expression.convert_to_operation()
            v = expression.value

            if not expression.is_operation:
                # short-circuit logic, return as early as possible
                if isinstance(expr.op, ast.And) and not v:
                    # TODO: return v, not True or False
                    # TODO: there can be many values if operations are present
                    value = False
                    break
                if isinstance(expr.op, ast.Or) and v:
                    value = True
                    break
                continue

            if value is None:
                value = v
                continue

            if isinstance(value, _Operation) or expression.is_operation:
                # change operator for operations
                py_op = self.operator_parser.py_operator({
                    ast.And: ast.BitAnd(),
                    ast.Or: ast.BitOr(),
                }[type(expr.op)])
            else:
                py_op = self.operator_parser.py_operator(expr.op)

            value = py_op(value, v)
        return Expression(value)

    def _convert_in_to_bool_op(self, expr: ast.Compare) -> Union[ast.Compare, ast.BoolOp]:
        """
        Convert expressions like:
            tick['X']     in [1, 2]   ----->   tick['X'] == 1  or tick['X'] == 2
            tick['X'] not in [1, 2]   ----->   tick['X'] != 1 and tick['X'] != 2
        """
        left, op, right = expr.left, expr.ops[0], expr.comparators[0]
        if not isinstance(op, (ast.In, ast.NotIn)):
            return expr

        assert len(expr.ops) == 1
        assert len(expr.comparators) == 1

        right_value = self.expression(right).value
        if isinstance(right_value, range) and right_value.step == 1:
            # replace tick['X'] in range(5, 10) to X >=5 AND X < 10
            return ast.BoolOp(
                op=ast.And(),
                values=[
                    ast.Compare(left=left, ops=[ast.GtE()], comparators=[ast.Constant(right_value.start)]),
                    ast.Compare(left=left, ops=[ast.Lt()], comparators=[ast.Constant(right_value.stop)]),
                ]
            )

        right_values = [ast.Constant(r) for r in right_value]

        bool_op: ast.boolop
        compare_op: ast.cmpop
        if isinstance(op, ast.In):
            bool_op, compare_op = ast.Or(), ast.Eq()
        else:
            bool_op, compare_op = ast.And(), ast.NotEq()

        values: List[ast.expr] = [
            ast.Compare(left=left, ops=[compare_op], comparators=[r])
            for r in right_values
        ]
        return ast.BoolOp(op=bool_op, values=values)

    def _convert_many_comparators_to_bool_op(self, expr: ast.Compare) -> Union[ast.Compare, ast.BoolOp]:
        """
        OneTick don't support compare expressions with many comparators
        so replacing them with several simple expressions.
        For example:
            1 < tick['X'] < 3,   ----->   tick['X'] > 1 AND tick['X'] < 3
        """
        if len(expr.comparators) == 1 and len(expr.ops) == 1:
            return expr

        comparators = []
        comparators.append(expr.left)
        ops = []
        for op, right in zip(expr.ops, expr.comparators):
            ops.append(op)
            comparators.append(right)

        bool_operands: List[ast.expr] = []
        for i in range(len(comparators) - 1):
            left, op, right = comparators[i], ops[i], comparators[i + 1]
            bool_operands.append(
                ast.Compare(left=left, ops=[op], comparators=[right])
            )
        return ast.BoolOp(op=ast.And(), values=bool_operands)

    def compare(self, expr: ast.Compare) -> Expression:
        """
        Compare operation expression: tick['X'] > 1, 1 < 2 < 3, tick['X'] in [1, 2] etc.
        """
        if len(expr.ops) > 1:
            return self.expression(
                self._convert_many_comparators_to_bool_op(expr)
            )

        op = expr.ops[0]
        if isinstance(op, (ast.In, ast.NotIn)):
            return self.expression(
                self._convert_in_to_bool_op(expr)
            )

        left = self.expression(expr.left)
        right = self.expression(expr.comparators[0])

        py_op = self.operator_parser.py_operator(op)
        value = py_op(left.value, right.value)
        return Expression(value)

    def keyword(self, expr: ast.keyword) -> Tuple[str, Any]:
        """
        Keyword argument expression from function call: func(key=value).
        Not converted to per tick script in any way, needed only in self.call() function.
        """
        arg = expr.arg
        val = self.expression(expr.value)
        assert arg is not None
        return arg, val.value

    def call(self, expr: ast.Call) -> Expression:
        """
        Any call expression, like otp.nsectime().
        The returned value of the call will be inserted in script.
        """
        func = self.expression(expr.func)

        # TODO: refactor, this code is copy-pasted from CaseExpressionParser.call()
        need_to_parse = False
        if not isinstance(func.value, types.BuiltinMethodType) and expr.args:
            node = expr.args[0]
            if isinstance(node, ast.Name) and node.id == self.fun.arg_name:
                # we will parse inner function call to OneTick per-tick script function
                # only if one of the function call arguments is 'tick' parameter of the original function
                need_to_parse = True

        if need_to_parse:
            err = None
            func_name = func.value.__name__
            if str(func.value) not in self.fun.emulator.FUNCTIONS:
                try:
                    fp = FunctionParser(func.value, emulator=self.fun.emulator,
                                        check_arg_name=self.fun.arg_name, inner_function=True)
                except Exception as e:
                    err = e
                else:
                    value = fp.per_tick_script()
                    self.fun.emulator.FUNCTIONS[str(func.value)] = (value, fp.args_annotations, fp.return_annotation)
            if err is None:
                _, args_annotations, return_type = self.fun.emulator.FUNCTIONS[str(func.value)]
                if expr.keywords:
                    raise ValueError(
                        f"Passing keyword parameters to per-tick script function '{func_name}' is not supported"
                    )
                args = []
                for arg, (name, dtype) in zip(expr.args[1:], args_annotations.items()):
                    if isinstance(arg, ast.Starred):
                        raise ValueError(
                            f"Passing starred parameter to per-tick script function '{func_name}' is not supported"
                        )
                    arg_expr = self.expression(arg)
                    msg = (f"In function '{func_name}' parameter '{name}'"
                           f" has type annotation '{dtype.__name__}',"
                           f" but the type of passed argument is '{arg_expr.dtype.__name__}'")
                    try:
                        widest_type = ott.get_type_by_objects([arg_expr.dtype, dtype])
                    except TypeError as e:
                        raise TypeError(msg) from e
                    if widest_type is not arg_expr.dtype:
                        raise TypeError(msg)
                    args.append(arg_expr)
                str_args = ', '.join(map(str, args))
                return Expression(_Operation(op_str=f'{func_name}({str_args})', dtype=return_type))

        args = []
        for arg in expr.args:
            # TODO: support starred in CaseExpressionParser.call()
            if isinstance(arg, ast.Starred):
                args.extend(self.expression(arg.value).value)
            else:
                args.append(self.expression(arg).value)
        keywords = dict(self.keyword(keyword) for keyword in expr.keywords)
        value = func.value(*args, **keywords)
        return Expression(value)

    def formatted_value(self, expr: ast.FormattedValue) -> Expression:
        """
        Block from the f-string in curly brackets, e.g.
            {tick['A']} and {123} in f"{tick['A']} {123}"
        """
        return self.expression(expr.value)

    def joined_str(self, expr: ast.JoinedStr) -> Expression:
        """
        F-string expression, like: f"{tick['A']} {123}"
        """
        expressions = [self.expression(value) for value in expr.values]
        value = None
        for expression in expressions:
            v = expression.value
            if expression.is_operation:
                v = v.apply(str)
            else:
                v = str(v)
            if value is None:
                value = v
            else:
                value = value + v
        return Expression(value)

    def list(self, expr: ast.List) -> Expression:
        """
        List expression, like: [1, 2, 3, 4, 5]
        """
        value = []
        for e in expr.elts:
            if isinstance(e, ast.Starred):
                value.extend(self.expression(e.value).value)
            else:
                value.append(self.expression(e).value)
        return Expression(value, expr=None)

    def tuple(self, expr: ast.Tuple) -> Expression:
        """
        Tuple expression, like: (1, 2, 3, 4, 5)
        """
        expression = self.list(expr)  # type: ignore[arg-type]
        expression.values = (tuple(expression.value),)
        return expression

    @property
    def _expression(self) -> dict:
        """Mapping from ast expression to parser functions"""
        mapping = {
            ast.Constant: self.constant,
            ast.Name: self.name,
            ast.Attribute: self.attribute,
            ast.Index: self.index,
            ast.Subscript: self.subscript,
            ast.BinOp: self.bin_op,
            ast.UnaryOp: self.unary_op,
            ast.BoolOp: self.bool_op,
            ast.Compare: self.compare,
            ast.Call: self.call,
            ast.FormattedValue: self.formatted_value,
            ast.JoinedStr: self.joined_str,
            ast.List: self.list,
            ast.Tuple: self.tuple,
            ast.Slice: self.slice,
        }
        deprecated = {
            'NameConstant': self.constant,
            'Str': self.string,
            'Num': self.number,
        }
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', DeprecationWarning)
            for name, callback in deprecated.items():
                if hasattr(ast, name):
                    mapping[getattr(ast, name)] = callback
        return mapping

    def expression(self, expr: ast.expr) -> Expression:
        """Return parsed expression according to its type."""
        return self._expression[type(expr)](expr)


class CaseExpressionParser(ExpressionParser):
    """
    Class with methods to convert ast expressions to CASE function.
    """
    def __init__(self, fun: 'FunctionParser'):
        super().__init__(fun)
        self.operator_parser = CaseOperatorParser()  # type: ignore[assignment]

    def _convert_bool_op_to_if_expr(self, expr: ast.expr) -> ast.expr:
        """
        Special case to convert bool operation to if expression.
        For example:
            lambda row: row['A'] or -1
        will be converted to:
            CASE(A != 0, 1, A, -1)
        """
        if not isinstance(expr, ast.BoolOp):
            return expr

        def get_if_expr(first, second):
            if isinstance(expr.op, ast.Or):
                return ast.IfExp(test=first, body=first, orelse=second)
            else:
                return ast.IfExp(test=first, body=second, orelse=first)

        first = None
        for i in range(len(expr.values) - 1):
            if first is None:
                first = expr.values[i]
                first = self._convert_bool_op_to_if_expr(first)
            second = expr.values[i + 1]
            second = self._convert_bool_op_to_if_expr(second)
            first = get_if_expr(first, second)
        assert first is not None
        return first

    def if_expr(self, expr: ast.IfExp) -> Expression:
        """
        If expression: 'A' if tick['X'] > 0 else 'B'.
        Do not confuse with if statement.
        Will be converted to OneTick case function: CASE(X > 0, 1, 'A', 'B').
        If condition value can be deduced before execution of script,
        then if or else value will be returned without using CASE() function.
        For example:
            tick['A'] if False else 3  -----------> 3
        """
        test = self.expression(expr.test)
        if test.predefined:
            # we can remove unnecessary branch if condition value is already known
            if test.value:
                return self.expression(expr.body)
            return self.expression(expr.orelse)
        body = self.expression(expr.body)
        orelse = self.expression(expr.orelse)
        test.convert_to_operation()

        str_expr = f'CASE({test}, 1, {body}, {orelse})'
        value = _LambdaIfElse(str_expr, ott.get_type_by_objects([*body.values, *orelse.values]))
        return Expression(value, expr=str_expr)

    def call(self, expr: ast.Call) -> Expression:
        """
        For CASE() function we support using inner functions that return valid case expression.
        """
        func = self.expression(expr.func)

        need_to_parse = False
        if not isinstance(func.value, types.BuiltinMethodType):
            for node in expr.args + [kw.value for kw in expr.keywords]:
                if isinstance(node, ast.Name) and node.id == self.fun.arg_name:
                    # we will parse inner function call to OneTick expression
                    # only if one of the function call arguments is
                    # 'tick' or 'row' parameter of the original function
                    need_to_parse = True
                    break
        orig_err = None
        if not need_to_parse:
            with _CompareTrackScope(emulation_enabled=False):
                try:
                    return super().call(expr)
                except Exception as err:
                    orig_err = err
                    uw = UserWarning(
                        f"Function '{astunparse(expr)}' can't be called in python, "
                        "will try to parse it to OneTick expression. "
                        f"Use '{self.fun.arg_name}' in function's signature to indicate "
                        "that this function can be parsed to OneTick expression."
                    )
                    uw.__cause__ = err
                    warnings.warn(uw)

        fp = FunctionParser(func.value, check_arg_name=False)

        kwargs = {}
        args = fp.ast_node.args.args
        if fp.is_method:
            args = args[1:]
        for arg, default in zip(reversed(args), reversed(fp.ast_node.args.defaults)):
            kwargs[arg.arg] = default
        kwargs.update({keyword.arg: keyword.value for keyword in expr.keywords})  # type: ignore[misc]
        for arg, arg_value in zip(args, expr.args):
            kwargs[arg.arg] = arg_value

        try:
            value = fp.compress()
        except Exception as err:
            try:
                return super().call(expr)
            except Exception:
                raise ValueError(
                    f"Can't convert function '{astunparse(expr)}' to CASE() expression."
                ) from err
        try:
            with self._replace_context(fp.closure_vars):
                # replace function parameters with calculated values
                value = fp.case_statement_parser._replace_nodes(value, replace_name=kwargs)
                return self.expression(value)
        except Exception as err:
            if orig_err is not None:
                raise err from orig_err
            raise err

    @property
    def _expression(self) -> dict:
        return dict(super()._expression.items() | {
            ast.IfExp: self.if_expr,
        }.items())


class CaseStatementParser:
    """
    Class with methods to convert ast statements to CASE function.
    """
    def __init__(self, fun: 'FunctionParser'):
        self.fun = fun
        self.expression_parser = CaseExpressionParser(fun=fun)
        self.operator_parser = CaseOperatorParser()

    T = TypeVar('T', bound=ast.AST)

    @staticmethod
    def _replace_nodes(node: T,
                       replace_name: Optional[Dict[str, ast.expr]] = None,
                       replace_break: Union[ast.stmt, Exception, Type[Exception], None] = None,
                       inplace: bool = False) -> T:
        """
        Function to replace expressions and statements inside ast.For node.

        Parameters
        ----------
        node
            ast node in which expressions and statements will be replaced
        inplace
            if True `node` object will be modified else it will be copied and the copy will be returned
        replace_name
            mapping from ast.Name ids to ast expressions.
            ast.Name nodes with these ids will be replaced with corresponding expressions.
        replace_break
            replace break statement with another statement.
            We can't execute for loop on real data here so we can't allow break statements at all.
            So we will replace them with statements from code after the for loop.
            If replace_break is Exception then exception will be raised when visiting ast.Break nodes.
        """
        class RewriteName(ast.NodeTransformer):
            def visit_Name(self, n: ast.Name):
                return (replace_name or {}).get(n.id) or n

            def visit_Continue(self, n: ast.Continue):
                # TODO: pass is not continue, we must allow only bodies with one statement in this case
                return ast.Pass()

            def visit_Break(self, n: ast.Break):
                if replace_break is None:
                    return n
                if inspect.isclass(replace_break) and issubclass(replace_break, Exception):
                    raise replace_break("Break is found in for loop and replacer is not provided")
                if isinstance(replace_break, Exception):
                    raise replace_break
                assert isinstance(replace_break, ast.stmt)
                return CaseStatementParser._replace_nodes(replace_break, replace_name=replace_name)

        if not inplace:
            node = deepcopy(node)
        RewriteName().visit(node)
        return node

    def _flatten_for_stmt(self,
                          stmt: ast.For,
                          replace_break: Union[ast.stmt, Exception, Type[Exception], None] = None,
                          stmt_after_for: Optional[ast.stmt] = None) -> List[ast.stmt]:
        """
        Convert for statement to list of copy-pasted statements from the body for each iteration.
        """
        stmts = []
        target = stmt.target
        assert isinstance(target, (ast.Name, ast.Tuple)), (
            f"Unsupported expression '{astunparse(target)}' is used in for statement."
            " Please, use variable or tuple of variables instead."
        )
        if isinstance(target, ast.Tuple):
            for t in target.elts:
                assert isinstance(t, ast.Name)
            targets = target.elts
        else:
            targets = [target]
        replace_name = {}
        iter_ = self.expression_parser.expression(stmt.iter)
        for iter_value in iter_.value:
            if not isinstance(iter_value, Iterable) or isinstance(iter_value, str):
                iter_value = [iter_value]
            replace_name = {
                target.id: ast.Constant(value)  # type: ignore[attr-defined]
                for target, value in zip(targets, iter_value)
            }
            for s in stmt.body:
                stmts.append(
                    self._replace_nodes(s,
                                        replace_name=replace_name,  # type: ignore[arg-type]
                                        replace_break=replace_break)
                )
        if stmt_after_for and replace_name:
            stmts.append(self._replace_nodes(stmt_after_for, replace_name=replace_name))  # type: ignore[arg-type]
        return stmts

    def _flatten_for_stmts(self, stmts: List[ast.stmt]) -> List[Union[ast.If, ast.Return, ast.Pass]]:
        """
        Find ast.For statements in list of statements and flatten them.
        Return list of statements without ast.For.
        Additionally raise exception if unsupported statement is found.
        """
        # TODO: support ast.For statements on deeper levels
        res_stmts = []
        for i, stmt in enumerate(stmts):
            if not isinstance(stmt, (ast.If, ast.Return, ast.For, ast.Pass)):
                raise ValueError(
                    "this function can't be converted to CASE function, "
                    "only for, if, return and pass statements are allowed"
                )
            if isinstance(stmt, ast.For):
                try:
                    res_stmts.extend(self._flatten_for_stmt(stmt, replace_break=ValueError))
                except ValueError:
                    stmts_left = len(stmts[i + 1:])
                    assert stmts_left in (0, 1), "Can't be more than one statement after break"
                    replace_break: Union[ast.Return, ast.Pass]
                    if stmts_left == 0:
                        stmt_after_for = None
                        replace_break = ast.Pass()
                    else:
                        stmt_after_for = stmts[i + 1]
                        assert isinstance(stmt_after_for, (ast.Return, ast.Pass)), (
                            'Can only use pass and return statements after for loop with break'
                        )
                        replace_break = stmt_after_for
                    res_stmts.extend(self._flatten_for_stmt(stmt,
                                                            replace_break=replace_break,
                                                            stmt_after_for=stmt_after_for))
                    break
            else:
                res_stmts.append(stmt)
        return res_stmts  # type: ignore[return-value]

    def _compress_stmts_to_one_stmt(self, stmts: List[ast.stmt], filler=None) -> Union[ast.If, ast.Return]:
        """
        List of if statements will be converted to one if statement.
        For example:
            if tick['X'] <= 1:
                if tick['X'] > 0:
                    return 1
                else:
                    pass
            else:
                if tick['X'] < 3:
                    return 2
            if tick['X'] <= 3:
                return 3
            return 4
        will be converted to:
            if tick['X'] <= 1:
                if tick['X'] > 0:
                    return 1
                else:
                    if tick['X'] <= 3:
                        return 3
                    else:
                        return 4
            else:
                if tick['X'] < 3:
                    return 2
                else:
                    if tick['X'] <= 3:
                        return 3
                    else:
                        return 4
        """
        filler = filler or ast.Pass()
        if not stmts:
            return filler
        stmt, *others = stmts
        if isinstance(stmt, ast.Return):
            return stmt
        if isinstance(stmt, ast.Pass):
            return filler
        filler = self._compress_stmts_to_one_stmt(others, filler=filler)
        if isinstance(stmt, ast.If):
            stmt.body = [self._compress_stmts_to_one_stmt(stmt.body, filler=filler)]
            if stmt.orelse:
                stmt.orelse = [self._compress_stmts_to_one_stmt(stmt.orelse, filler=filler)]
            elif filler:
                stmt.orelse = [filler]
            assert stmt.orelse
            return stmt
        raise ValueError(
            "this function can't be converted to CASE function, "
            "only for, if, return and pass statements are allowed"
        )

    def _replace_local_variables(self, stmts: List[ast.stmt]) -> List[ast.stmt]:
        """
        We support local variables only by calculating their value and
        replacing all it's occurrences in the code after variable definition.
        For example:
            a = 12345
            if a:
                return a
            return 0
        will be converted to:
            if 12345:
                return 12345
            return 0
        """
        replace_name = {}
        res_stmts = []
        for stmt in stmts:
            if isinstance(stmt, ast.Assign):
                assert len(stmt.targets) == 1, 'Unpacking local variables is not supported yet'
                var, val = stmt.targets[0], stmt.value
                assert isinstance(var, ast.Name)
                replace_name[var.id] = val
                continue
            res_stmts.append(
                self._replace_nodes(stmt, replace_name=replace_name)
            )
        return res_stmts

    def if_stmt(self, stmt: ast.If) -> ast.IfExp:
        """
        Classic if statement with limited set of allowed statements in the body:
          * only one statement in the body
          * statement can be return or another if statement with same rules as above

        For example:
            if tick['X'] > 0:
                return 'POS'
            elif tick['X'] == 0:
                return 'ZERO'
            else:
                return 'NEG'
        will be converted to OneTick's CASE function:
            CASE(X > 0, 1, 'POS', CASE(X = 0, 1, 'ZERO', 'NEG'))
        """
        # TODO: support many statements in body
        if len(stmt.body) != 1:
            raise ValueError("this function can't be converted to CASE function, "
                             "too many statements in if body")
        body = self.statement(stmt.body[0])

        if len(stmt.orelse) > 1:
            raise ValueError("this function can't be converted to CASE function, "
                             "too many statements in else body")
        if stmt.orelse and not isinstance(stmt.orelse[0], ast.Pass):
            orelse = self.statement(stmt.orelse[0])
        else:
            e = self.expression_parser.expression(body)
            orelse = ast.Constant(ott.default_by_type(ott.get_type_by_objects(e.values)))
        return ast.IfExp(test=stmt.test, body=body, orelse=orelse)

    def return_stmt(self, stmt: ast.Return) -> ast.expr:
        """
        Return statement.
        Will be converted to value according to OneTick's syntax.
        """
        if stmt.value is None:
            raise ValueError('return statement must have value when converting to CASE function')
        return stmt.value

    def pass_stmt(self, _stmt: ast.Pass) -> ast.Constant:
        """
        Pass statement.
        Will be converted to None according to OneTick's syntax.
        """
        return ast.Constant(None)

    def compress(self, stmts: List[ast.stmt]) -> Union[ast.If, ast.Return]:
        """
        Compress list of statements to single statement.
        This is possible only if simple if and return statements are used.
        """
        stmts = self._replace_local_variables(stmts)
        flat_stmts = self._flatten_for_stmts(stmts)
        stmt = self._compress_stmts_to_one_stmt(flat_stmts)  # type: ignore[arg-type]
        return stmt

    def statement(self, stmt: ast.stmt) -> ast.expr:
        """Return statement converted to expression."""
        return {
            ast.If: self.if_stmt,
            ast.Return: self.return_stmt,
            ast.Pass: self.pass_stmt,
        }[type(stmt)](stmt)  # type: ignore[operator]


class StatementParser(CaseStatementParser):
    """
    Class with methods to convert ast statements to per tick script lines.
    """

    def __init__(self, fun: 'FunctionParser'):
        super().__init__(fun)
        self.expression_parser = ExpressionParser(fun=fun)  # type: ignore[assignment]
        self.operator_parser = OperatorParser()
        self._for_counter = 0

    @staticmethod
    def _transform_if_expr_to_if_stmt(stmt: Union[ast.Assign, ast.AugAssign]) -> ast.If:
        """
        Per tick script do not support if expressions, so converting it to if statement.
        For example:
            tick['X'] = 'A' if tick['S'] > 0 else 'B'
        will be converted to:
            if (S > 0) {
                X = 'A';
            }
            else {
                X = 'B';
            }
        """
        if not isinstance(stmt.value, ast.IfExp):
            raise ValueError()
        if_expr: ast.IfExp = stmt.value
        body, orelse = deepcopy(stmt), deepcopy(stmt)
        body.value = if_expr.body
        orelse.value = if_expr.orelse

        return ast.If(
            test=if_expr.test,
            body=[body],
            orelse=[orelse],
        )

    def assign(self, stmt: ast.Assign) -> str:
        """
        Assign statement: tick['X'] = 1
        Will be converted to OneTick syntax: X = 1;
        """
        assert len(stmt.targets) == 1, 'Unpacking variables is not yet supported'

        if isinstance(stmt.value, ast.IfExp):
            if_stmt = self._transform_if_expr_to_if_stmt(stmt)
            return self.statement(if_stmt)

        var = self.expression_parser.expression(stmt.targets[0])
        val = self.expression_parser.expression(stmt.value)

        default_expr = f'{var} = {val};'

        if var.lhs:
            expr = var.value(val.value)
            return expr or default_expr

        if var.is_local_variable:
            var_name = var.value.name
            if val.is_static:
                val = Expression(val.value.value)
                if var_name in self.fun.emulator.STATIC_VARS:
                    raise ValueError(f"Trying to define static variable '{var_name}' more than once")
                if var_name in self.fun.emulator.LOCAL_VARS:
                    raise ValueError(f"Can't redefine variable '{var_name}' as static")

                if self.fun.emulator.NEW_VALUES:
                    raise ValueError('Mixed definition of static variables and new columns is not supported')

                if val.is_tick:
                    # recreating tick object here, because it doesn't have name yet
                    self.fun.emulator.STATIC_VARS[var_name] = val.dtype(var_name)
                    return f'static {val.value._definition} {var};'
                self.fun.emulator.STATIC_VARS[var_name] = val.value
                return f'static {ott.type2str(val.dtype)} {var} = {val};'

            variables = None
            if var_name in self.fun.emulator.STATIC_VARS:
                variables = self.fun.emulator.STATIC_VARS
            elif var_name in self.fun.emulator.LOCAL_VARS:
                variables = self.fun.emulator.LOCAL_VARS

            if variables is None:
                if val.is_tick:
                    raise ValueError('Only primitive types are allowed for non static local variables.')
                if self.fun.emulator.NEW_VALUES:
                    raise ValueError('Mixed definition of local variables and new columns is not supported')
                self.fun.emulator.LOCAL_VARS[var_name] = val.value
                return f'{ott.type2str(val.dtype)} {var} = {val};'

            dtype = ott.get_type_by_objects([variables[var_name]])
            if val.dtype != dtype:
                raise ValueError(f"Wrong type for variable '{var_name}': should be {dtype}, got {val.dtype}")

        return default_expr

    def aug_assign(self, stmt: ast.AugAssign) -> str:
        """
        Assign with inplace operation statement: tick['X'] += 1.
        Will be converted to OneTick syntax: X = X + 1;
        """
        target = deepcopy(stmt.target)
        target.ctx = ast.Load()
        return self.assign(
            ast.Assign(
                targets=[stmt.target],
                value=ast.BinOp(
                    left=target,
                    op=stmt.op,
                    right=stmt.value,
                )
            )
        )

    def if_stmt(self, stmt: ast.If) -> str:  # type: ignore[override]
        """
        Classic if statement:
            if tick['X'] > 0:
                tick['Y'] = 1
            elif tick['X'] == 0:
                tick['Y'] = 0
            else:
                tick['Y'] = -1
        Will be converted to:
            if (X > 0) {
                Y = 1;
            }
            else {
                if (X == 0) {
                    Y = 0;
                }
                else {
                    Y = -1;
                }
            }
        """
        test = self.expression_parser.expression(stmt.test)
        test.convert_to_operation()
        body = [self.statement(s) for s in stmt.body]
        orelse = [self.statement(s) for s in stmt.orelse]
        if test.predefined:
            if test.value:
                return '\n'.join(body)
            return '\n'.join(orelse)
        lines = []
        lines.append('if (%s) {' % test)
        lines.extend(body)
        lines.append('}')
        if orelse:
            lines.append('else {')
            lines.extend(orelse)
            lines.append('}')
        return '\n'.join(lines)

    def return_stmt(self, stmt: ast.Return) -> str:  # type: ignore[override]
        """
        Return statement.
        For now we support returning only boolean values or nothing.
        Will be converted to: return true;
        """
        # if return is empty then it is not filter
        v = stmt.value if stmt.value is not None else ast.Constant(value=True)
        value = self.expression_parser.expression(v)
        dtype = ott.get_object_type(value.value)
        if not self.fun.inner_function:
            if dtype is not bool:
                raise TypeError(f"Not supported return type {dtype}")
            if stmt.value is not None:
                self.fun.returns = True
        else:
            assert isinstance(self.fun.ast_node, ast.FunctionDef)
            msg = (f"Function '{self.fun.ast_node.name}'"
                   f" has return annotation '{self.fun.return_annotation.__name__}',"
                   f" but the type of statement ({astunparse(stmt)}) is '{dtype.__name__}'")
            try:
                widest_type = ott.get_type_by_objects([dtype, self.fun.return_annotation])
            except TypeError as e:
                raise TypeError(msg) from e
            if widest_type is not self.fun.return_annotation:
                raise TypeError(msg)
            self.fun.returns = True
        return f'return {value};'

    @staticmethod
    def _check_break(*nodes) -> bool:
        """
        Check if break statement in the list of nodes (recursively).
        """
        class FoundBreakException(Exception):
            pass

        class FindBreak(ast.NodeVisitor):
            def visit_Break(self, n: ast.Break):
                raise FoundBreakException()

        try:
            for node in nodes:
                FindBreak().visit(node)
            return False
        except FoundBreakException:
            return True

    def while_stmt(self, stmt: ast.While) -> str:
        """
        Classic while statement:
            while tick['X'] > 0:
                tick['Y'] = 1
        Will be converted to:
            while (X > 0) {
                Y = 1;
            }
        """
        test = self.expression_parser.expression(stmt.test)
        test.convert_to_operation()
        body = [self.statement(s) for s in stmt.body]
        if test.predefined:
            is_break_found = self._check_break(*stmt.body)
            if not is_break_found:
                raise ValueError(f'The condition of while statement always evaluates to {bool(test.value)}'
                                 ' and there is no break statement in the loop body.'
                                 ' That will result in infinite loop.'
                                 ' Change condition or add break statements.')
        lines = []
        lines.append('while (%s) {' % test)
        lines.extend(body)
        lines.append('}')
        return '\n'.join(lines)

    def for_stmt(self, stmt: ast.For) -> str:
        """
        For now for statement in most cases will not be converted to per tick script's for statement.
        Instead, the statements from the body of the for statement will be duplicated
        for each iteration.

        For example:
            for i in [1, 2, 3]:
                tick['X'] += i
        will be converted to:
            X += 1;
            X += 2;
            X += 3;

        But simple case with range object will be translated more correctly:
        For example:
            for i in range(1, 4):
                tick['X'] += i
        will be converted to:
            for (LOCAL::i = 1; LOCAL::i < 4; LOCAL::i += 1) {
                X += LOCAL::i;
            }
        """
        lines = []
        iter_ = self.expression_parser.expression(stmt.iter)
        if isinstance(iter_.value, _TickSequence):
            target = stmt.target
            assert isinstance(target, ast.Name), "Tuples can't be used while iterating on tick sequences"
            state_tick = iter_.value._tick_obj(target.id)
            # TODO: ugly
            state_tick_name = f"_______state_tick_{self._for_counter}_______"
            self._for_counter += 1
            ast_tick = ast.Name(state_tick_name, ctx=ast.Load())
            lines.append('for (%s %s : %s) {' % (state_tick._definition, state_tick, iter_.value))
            with self.expression_parser._replace_context(
                inspect.ClosureVars({}, {state_tick_name: state_tick}, {}, set())
            ):
                for s in stmt.body:
                    s = self._replace_nodes(s, replace_name={target.id: ast_tick})
                    lines.append(self.statement(s))
            lines.append('}')
        elif isinstance(iter_.value, range):
            target = stmt.target
            assert isinstance(target, ast.Name), "Tuples can't be used in for loop"
            var_name = target.id
            if var_name not in self.fun.emulator.LOCAL_VARS:
                # initialize counter variable
                self.fun.emulator.LOCAL_VARS[var_name] = int
                self.fun.emulator.LOCAL_VARS_NEW_VALUES[var_name].append(0)
            elif self.fun.emulator.LOCAL_VARS[var_name] is not int:
                raise ValueError(f'Variable {var_name} was declared before with conflicting type.')
            counter_var = LocalVariable(var_name, int)
            counter_var_str = str(counter_var)
            range_obj = iter_.value
            start_expr = f'{counter_var_str} = {range_obj.start}'
            if (range_obj.start < range_obj.stop and range_obj.step <= 0 or
                    range_obj.start > range_obj.stop and range_obj.step >= 0):
                raise ValueError(f'Range object {range_obj} will result in infinite loop')
            if range_obj.start < range_obj.stop:
                condition_expr = f'{counter_var_str} < {range_obj.stop}'
            else:
                condition_expr = f'{counter_var_str} > {range_obj.stop}'
            increment_expr = f'{counter_var_str} += {range_obj.step}'
            lines.append('for (%s; %s; %s) {' % (start_expr, condition_expr, increment_expr))
            try:
                for s in stmt.body:
                    lines.append(self.statement(s))
                lines.append('}')
            except Exception:
                if var_name in self.fun.emulator.LOCAL_VARS_NEW_VALUES:
                    self.fun.emulator.LOCAL_VARS_NEW_VALUES.pop(var_name)
                    self.fun.emulator.LOCAL_VARS.pop(var_name)
                lines = [self.statement(s) for s in self._flatten_for_stmt(stmt)]
        else:
            lines = [self.statement(s) for s in self._flatten_for_stmt(stmt)]
        return '\n'.join(lines)

    def break_stmt(self, _stmt: ast.Break) -> str:
        return 'break;'

    def continue_stmt(self, _stmt: ast.Continue) -> str:
        return 'continue;'

    def pass_stmt(self, stmt: ast.Pass) -> str:  # type: ignore[override]
        """Pass statement is not converted to anything"""
        return ''

    def yield_expr(self, expr: ast.Yield) -> Expression:
        """
        Yield expression, like: yield
        Values for yield are not supported.
        Will be translated to PROPAGATE_TICK() function.
        Can be used only as a statement, so this function is here and not in ExpressionParser.
        """
        if expr.value is not None:
            raise ValueError("Passing value with yield expression is not supported.")
        return Expression('PROPAGATE_TICK();')

    def expression(self, stmt: ast.Expr) -> str:
        """
        Here goes raw strings and yield expression.
        For example:
            if tick['A'] == 0:
                'return 0;'
        Here 'return 0;' is used as a statement and an expression.
        Expression's returned value *must* be a string and
        this string will be injected in per tick script directly.
        """
        if isinstance(stmt.value, ast.Yield):
            expression = self.yield_expr(stmt.value)
        else:
            expression = self.expression_parser.expression(stmt.value)
        assert isinstance(expression.value, (str, _Operation)), (
            f"The statement '{astunparse(stmt)}' can't be used here"
            " because the value of such statement can be string only"
            " as it's value will be injected directly in per tick script."
        )
        value = str(expression.value)
        if value and value[-1] != ';':
            value += ';'
        return value

    def with_stmt(self, stmt: ast.With) -> str:
        """
        Used only with special context managers. Currently only `_ONCE` is supported.
        """
        if len(stmt.items) != 1:
            raise ValueError('Currently it is possible to use only one context manager in single with statement')
        with_item = stmt.items[0]
        if with_item.optional_vars:
            raise ValueError('It is not allowed to use "as" in with statements for per-tick script')
        context_expr = with_item.context_expr
        if isinstance(context_expr, ast.Call):
            expr = self.expression_parser.expression(context_expr.func)
        else:
            raise ValueError(f'{context_expr} is not called')
        if not issubclass(expr.value, once):
            raise ValueError(f'{expr.value} is not supported in per-tick script with statements')
        return expr.value().get_str('\n'.join([self.statement(s) for s in stmt.body]))

    def statement(self, stmt: ast.stmt) -> str:  # type: ignore[override]
        """Return parsed statement according to its type."""
        return {
            ast.Assign: self.assign,
            ast.AugAssign: self.aug_assign,
            ast.If: self.if_stmt,
            ast.Return: self.return_stmt,
            ast.While: self.while_stmt,
            ast.For: self.for_stmt,
            ast.Break: self.break_stmt,
            ast.Continue: self.continue_stmt,
            ast.Pass: self.pass_stmt,
            ast.Expr: self.expression,
            ast.With: self.with_stmt,
        }[type(stmt)](stmt)  # type: ignore[operator]


class EndOfBlock(Exception):
    pass


class LambdaBlockFinder:
    """
    This is simplified version of
    inspect.BlockFinder
    that supports multiline lambdas.
    """

    OPENING_BRACKETS = {
        '[': ']',
        '(': ')',
        '{': '}',
    }
    CLOSING_BRACKETS = {c: o for o, c in OPENING_BRACKETS.items()}
    BRACKETS_MATCHING = dict(OPENING_BRACKETS.items() | CLOSING_BRACKETS.items())

    def __init__(self):
        # current indentation level
        self.indent = 0
        # row and column index for the start of lambda expression
        self.start = None
        # row and column index for the end of lambda expression
        self.end = None
        # stack with brackets
        self.brackets = deque()
        self.prev = None
        self.current = None

    def tokeneater(self, type, token, srowcol, erowcol, line, start_row=0):
        srowcol = (srowcol[0] + start_row, srowcol[1])
        erowcol = (erowcol[0] + start_row, erowcol[1])
        self.prev = self.current
        self.current = tokenize.TokenInfo(type, token, srowcol, erowcol, line)
        self.end = erowcol
        if token == 'lambda':
            self.start = srowcol
        elif type == tokenize.INDENT:
            self.indent += 1
        elif type == tokenize.DEDENT:
            self.indent -= 1
            # the end of matching indent/dedent pairs ends a block
            if self.indent <= 0:
                raise EndOfBlock
        elif not self.start:
            self.indent = 0
        elif type == tokenize.NEWLINE:
            if self.indent == 0 or (
                # if lambda is the argument of the function
                self.prev and self.prev.type == tokenize.OP and self.prev.string == ','
            ):
                raise EndOfBlock
        elif token in self.OPENING_BRACKETS:
            self.brackets.append(token)
        elif token in self.CLOSING_BRACKETS:
            try:
                assert self.brackets.pop() == self.CLOSING_BRACKETS[token]
            except (IndexError, AssertionError):
                self.end = self.prev.end
                raise EndOfBlock  # noqa: W0707


def get_lambda_source(lines):
    """Extract the block of lambda code at the top of the given list of lines."""
    blockfinder = LambdaBlockFinder()
    start_row = 0
    while True:
        try:
            tokens = tokenize.generate_tokens(iter(lines[start_row:]).__next__)
            for _token in tokens:
                blockfinder.tokeneater(*_token, start_row=start_row)
            break
        except IndentationError as e:
            # indentation errors are possible because
            # we started eating tokens from line with lambda
            # not from the start of the statement
            # trying to eat again from the current row in this case
            start_row = e.args[1][1] - 1
            continue
        except EndOfBlock:
            break
    start_row, start_column = blockfinder.start
    end_row, end_column = blockfinder.end
    # crop block to get rid of tokens from the context around lambda
    lines = lines[start_row - 1: end_row]
    lines[-1] = lines[-1][:end_column]
    lines[0] = lines[0][start_column:]
    # add brackets around lambda in case it is multiline lambda
    return ''.join(['(', *lines, ')'])


def is_lambda(lambda_f) -> bool:
    return isinstance(lambda_f, types.LambdaType) and lambda_f.__name__ == '<lambda>'


def get_source(lambda_f) -> str:
    """
    Get source code of the function or lambda.
    """
    if is_lambda(lambda_f):
        # that's a hack for multiline lambdas in brackets
        # inspect.getsource parse them wrong
        source_lines, lineno = inspect.findsource(lambda_f)
        if 'lambda' not in source_lines[lineno]:
            # inspect.findsource fails sometimes too
            lineno = lambda_f.__code__.co_firstlineno + 1
            while 'lambda' not in source_lines[lineno]:
                lineno -= 1
        source = get_lambda_source(source_lines[lineno:])
    else:
        source = inspect.getsource(lambda_f)
    # doing dedent because self.ast_node do not like indented source code
    return textwrap.dedent(source)


class FunctionParser:
    """
    Class to parse callable objects (lambdas and functions) to
    OneTick's per tick script or case functions.
    Only simple functions corresponding to OneTick syntax supported
    (without inner functions, importing modules, etc.)
    You can call simple functions inside,
    do operations with captured variables (without assigning to them),
    but using non-pure functions is not recommended because
    the code in function may not be executed in the order you expect.
    """
    SOURCE_CODE_ATTRIBUTE = '___SOURCE_CODE___'

    def __init__(self, lambda_f, emulator=None, check_arg_name=True, inner_function=False):
        """
        Parameters
        ----------
        emulator
            otp.Source emulator that will be tracking changes made to source
        check_arg_name
            if True, only callables with zero or one parameter will be allowed.
        inner_function
            if True, then function is treated like inner per-tick script function.
            First argument will be checked and more arguments will be allowed too.
        """

        assert isinstance(lambda_f, (types.LambdaType, types.FunctionType, types.MethodType)), (
            f"It is expected to get a function, method or lambda, but got '{type(lambda_f)}'"
        )
        self.lambda_f = lambda_f
        self.emulator = emulator
        self.check_arg_name = check_arg_name
        self.inner_function = inner_function
        if self.inner_function and not self.check_arg_name:
            self.check_arg_name = self.check_arg_name or True
        self.statement_parser = StatementParser(fun=self)
        self.expression_parser = ExpressionParser(fun=self)
        self.case_expression_parser = CaseExpressionParser(fun=self)
        self.case_statement_parser = CaseStatementParser(fun=self)
        # if the function returns some values or not
        self.returns = False
        # will be set to True when args_annotations will be calculated, need it to break recursion
        self._from_args_annotations = False
        # calling property here, so we can raise exception as early as possible
        _ = self.arg_name

    @cached_property
    def is_method(self) -> bool:
        return isinstance(self.lambda_f, types.MethodType)

    @cached_property
    def source_code(self) -> str:
        """
        Get source code of the function or lambda.
        """
        # first try to get code from special attribute else get code the usual way
        return getattr(self.lambda_f, self.SOURCE_CODE_ATTRIBUTE, None) or get_source(self.lambda_f)

    @cached_property
    def closure_vars(self) -> inspect.ClosureVars:
        """
        Get closure variables of the function.
        These are variables that were captured from the context before function definition.
        For example:
            A = 12345
            def a():
                print(A + 1)
        In this function variable A is the captured variable.
        We need closure variables, so we can use them when parsing ast tree.
        """
        return inspect.getclosurevars(self.lambda_f)

    @cached_property
    def ast_node(self) -> Union[ast.FunctionDef, ast.Lambda]:
        """
        Convert function or lambda to ast module statement.
        """
        source_code = self.source_code
        tree = ast.parse(source_code)
        for node in ast.walk(tree):
            if isinstance(node, (ast.FunctionDef, ast.Lambda)):
                if isinstance(node, ast.FunctionDef) and ast.get_docstring(node):
                    # remove comment section from function body
                    node.body.pop(0)
                return node
        raise ValueError("Can't find function or lambda in source code")

    @cached_property
    def arg_name(self) -> Optional[str]:
        """Get name of the first function or lambda argument."""
        node = self.ast_node
        argv = list(node.args.args)
        argc = len(argv)
        if argc > 1 and argv[0].arg == 'self' and self.is_method:
            argv.pop(0)
            argc -= 1
        if self.check_arg_name and argc > 1 and not self.inner_function:
            raise ValueError(
                "It is allowed to pass only functions or lambdas that take either one or"
                f" zero parameters, but got {argc}"
            )
        arg_name = argv[0].arg if argv else None
        if isinstance(self.check_arg_name, str) and arg_name != self.check_arg_name:
            assert isinstance(node, ast.FunctionDef)
            msg = f"Function '{node.name}' is expected to have first argument named '{self.check_arg_name}'"
            if arg_name is None:
                msg += ', but no argument is found'
            else:
                msg += f", but argument with name '{arg_name}' is found"
            raise ValueError(msg)
        return arg_name

    @cached_property
    def args_annotations(self) -> dict:
        node = self.ast_node
        if not isinstance(node, ast.FunctionDef):
            return {}
        argv = list(node.args.args)
        if argv and argv[0].arg == 'self' and self.is_method:
            argv.pop(0)
        if argv and argv[0].arg == self.arg_name:
            argv.pop(0)
        if node.args.defaults:
            raise ValueError("Default values for arguments are not supported"
                             f" in per-tick script function '{node.name}'")
        annotations = {}
        for arg in argv:
            name = arg.arg
            annotation = getattr(arg, 'annotation', None)
            if not annotation:
                raise ValueError(f"Parameter '{name}' in function '{node.name}' doesn't have type annotation")
            # TODO: remove hacking
            self._from_args_annotations = True
            dtype = self.expression_parser.expression(annotation).value
            self._from_args_annotations = False
            if not ott.is_type_basic(dtype):
                raise ValueError(f"Parameter '{name}' in function '{node.name}' has unsupported type: {dtype}")
            annotations[name] = dtype

        return annotations

    @cached_property
    def return_annotation(self):
        node = self.ast_node
        annotation = getattr(node, 'returns')
        if not annotation:
            raise ValueError(f"Function '{node.name}' doesn't have return type annotation")
        dtype = self.expression_parser.expression(annotation).value
        if not ott.is_type_basic(dtype):
            raise ValueError(f"Function '{node.name}' has unsupported return type: {dtype}")
        return dtype

    def per_tick_script(self) -> str:
        """
        Convert function to OneTick's per tick script.
        """
        node = self.ast_node

        lines = []

        assert isinstance(node, ast.FunctionDef), 'lambdas are not supported in per-tick-script yet'
        function_def: ast.FunctionDef = node

        for stmt in function_def.body:
            line = self.statement_parser.statement(stmt)
            if line:
                lines.append(line)

        if not self.inner_function and self.returns:
            # if there were return statement anywhere in the code
            # then we add default return at the end
            # TODO: but the default behaviour in OneTick is to propagate all ticks?
            #       changing that will break backward-compatibility
            lines.append(self.statement_parser.statement(ast.Return(ast.Constant(False))))

        if self.emulator is not None and not self.inner_function:
            # per tick script syntax demand that we declare variables before using them
            # so we get all new variables from emulator and declare them.

            def var_definition(key, values):
                dtype = ott.get_type_by_objects(values)
                return f'{ott.type2str(dtype)} {str(key)} = {ott.value2str(ott.default_by_type(dtype))};'

            new_columns = []
            for key, values in self.emulator.NEW_VALUES.items():
                new_columns.append(var_definition(key, values))
            new_local_vars = []
            for key, values in self.emulator.LOCAL_VARS_NEW_VALUES.items():
                new_local_vars.append(var_definition(LocalVariable(key), values))
            lines = new_columns + new_local_vars + lines

        if not lines:
            raise ValueError("The resulted body of PER TICK SCRIPT is empty")

        if not self.inner_function:
            lines = ['long main() {'] + lines + ['}']
            for function, *_ in self.emulator.FUNCTIONS.values():
                lines.append(function)
        else:
            if not self.returns:
                raise ValueError(f"Function '{node.name}' must return values")
            return_type = ott.type2str(self.return_annotation)
            args = [
                f'{ott.type2str(dtype)} {name}'
                for name, dtype in self.args_annotations.items()
            ]
            args: str = ', '.join(args)  # type: ignore[no-redef]
            lines = [f'{return_type} {node.name}({args})' + ' {'] + lines + ['}']

        return '\n'.join(lines) + '\n'

    def compress(self) -> ast.expr:
        """
        Convert lambda or function to AST expression.
        """
        node = self.ast_node
        if isinstance(node, ast.Lambda):
            return node.body
        stmt = self.case_statement_parser.compress(node.body)
        return self.case_statement_parser.statement(stmt)

    def case(self) -> Tuple[str, List[Any]]:
        """
        Convert lambda or function to OneTick's CASE() function.
        """
        expr = self.compress()
        expr = self.case_expression_parser._convert_bool_op_to_if_expr(expr)
        expression = self.case_expression_parser.expression(expr)
        # this will raise type error if type of the expression is not supported
        ott.default_by_type(ott.get_type_by_objects(expression.values))
        return str(expression), expression.values


def remote(fun):
    """
    This decorator is needed in case function ``fun``
    is used in :py:meth:`~onetick.py.Source.apply` method in a `Remote OTP with Ray` context.

    We want to get source code of the function locally
    because we will not be able to get source code on the remote server.

    See also
    --------
    :ref:`Remote OTP with Ray <ray-remote>`.
    """
    # see PY-424
    @wraps(fun)
    def wrapper(*args, **kwargs):
        return fun(*args, **kwargs)
    setattr(wrapper, FunctionParser.SOURCE_CODE_ATTRIBUTE, get_source(fun))
    return wrapper


class once:
    """
    Used with a statement or a code block to make it run only once (the first time control reaches to the statement).
    """
    def __enter__(self):
        # __enter__ and __exit__ methods are only used to express syntax for per-tick script, thus no implementation
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        # __enter__ and __exit__ methods are only used to express syntax for per-tick script, thus no implementation
        pass

    def get_str(self, string: str) -> str:
        return f"_ONCE\n{{\n{string}\n}}"


class Once(once):
    def __init__(self):
        warnings.warn('Using `otp.Once` is deprecated, please, use `otp.once` instead', FutureWarning)
        super().__init__()


def logf(message, severity, *args) -> str:
    """
    Call built-in OneTick ``LOGF`` function from per-tick script.

    Parameters
    ----------
    message: str
        Log message/format string. The underlying formatting engine is the Boost Format Library:
        https://www.boost.org/doc/libs/1_53_0/libs/format/doc/format.html

    severity: str
        Severity of message. Supported values: ``ERROR``, ``WARNING`` and ``INFO``.

    args: list
        Parameters for format string (optional).

    Returns
    -------
    str

    Examples
    --------
    >>> t = otp.Ticks({'X': [1, 2, 3]})

    >>> def test_script(tick):
    ...     otp.logf("Tick with value X=%1% processed", "INFO", tick["X"])

    >>> t = t.script(test_script)

    See also
    --------
    :ref:`Per-Tick Script Guide <python callable parser>`
    """

    if severity not in {"ERROR", "WARNING", "INFO"}:
        raise ValueError(f"Param severity expected to be one of ERROR, WARNING or INFO. Got \"{severity}\"")

    message = ott.value2str(message)
    severity = ott.value2str(severity)

    if args:
        params = ", ".join([ott.value2str(arg) for arg in args])
        return f"LOGF({message}, {severity}, {params});"
    else:
        return f"LOGF({message}, {severity});"


def throw_exception(message: str) -> str:
    """
    Call built-in OneTick ``THROW_EXCEPTION`` function from per-tick script.

    Parameters
    ----------
    message: str
        Message string that defines the error message to be thrown.

    Returns
    -------
    str

    Examples
    --------
    >>> t = otp.Ticks({'X': [1, -2, 6]})

    >>> def test_script(tick):
    ...     if tick["X"] <= 0:
    ...         otp.throw_exception("Tick column X should be greater than zero.")

    >>> t = t.script(test_script)

    See also
    --------
    :ref:`Per-Tick Script Guide <python callable parser>`
    """
    message = ott.value2str(message)
    return f"THROW_EXCEPTION({message});"



from typing import Optional, Union


class RemoteTS:
    PROTOCOLS = ('ws', 'wss', 'http')

    def __init__(
        self,
        host: Union[str, 'LoadBalancing', 'FaultTolerance'],
        port: Optional[Union[int, str]] = None,
        protocol: Optional[str] = None,
        resource: Optional[str] = None,
        cep: bool = False,
    ):
        """
        Class representing remote tick-server.
        Can be :py:meth:`used <onetick.py.Session.use>`
        in :py:class:`~onetick.py.Session` as well as the local databases.

        Parameters
        ----------
        host: str, :py:class:`~onetick.py.servers.LoadBalancing`, :py:class:`~onetick.py.servers.FaultTolerance`
            In case of string: string with this format: ``[proto://]hostname:port[/resource]``.
            Parameters in square brackets are optional.

            Otherwise, configuration of LoadBalancing and/or FaultTolerance (please, check corresponding classes)
        port: int, str, optional
            The port number of the remote tick-server.
            If not specified here, can be specified in the ``host`` parameter.
        protocol: str, optional
            The protocol to connect with.
            If not specified here, can be specified in the ``host`` parameter.
        resource: str, optional
            The resource of the host.
            If not specified here, can be specified in the ``host`` parameter.
        cep: bool
            Specifies if the remote server is the CEP-mode tick-server.

        Examples
        --------

        Specify host name and port together in first parameter:

        >>> session.use(otp.RemoteTS('server.onetick.com:50015'))  # doctest: +SKIP

        Additionally specify websocket protocol and resource of the remote server:

        >>> session.use(otp.RemoteTS('wss://data.onetick.com:443/omdwebapi/websocket'))  # doctest: +SKIP

        Combination of LoadBalancing and FaultTolerance can be used for host parameter:

        >>> RemoteTS(FaultTolerance(LoadBalancing('host1:4001', 'host2:4002'),
        ...                         LoadBalancing('host3:4003', 'host3:4004')) # doctest: +SKIP
        """
        self.cep = cep

        self._protocol = protocol
        self._host = host
        self._port = port
        self._resource = resource

        if not isinstance(self._host, str):
            if self._port or self._protocol or self._resource:
                raise ValueError('When host is not a string, other parameters cannot be specified')
            return

        protocol, host, port, resource = self._parse_url(host)
        if any([protocol, port, resource]) and any([self._protocol, self._port, self._resource]):
            raise ValueError('Mixed url parts in `host` parameter and in other parameters')

        self._protocol = self._protocol or protocol
        self._host = host
        self._port = self._port or port
        self._resource = self._resource or resource

        if self._protocol and self._protocol not in self.PROTOCOLS:
            raise ValueError(f"Protocol '{self._protocol}' is not supported")

        if not self._port:
            raise ValueError('Port is not set')

    def _parse_url(self, url):
        proto = None
        if '://' in url:
            proto, _, url = url.partition('://')
        resource = None
        if '/' in url:
            url, _, resource = url.partition('/')
        port = None
        if ':' in url:
            url, _, port = url.partition(':')
        hostname = url
        return proto, hostname, port, resource

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if isinstance(self._host, (FaultTolerance, LoadBalancing)):
            return str(self._host)
        url = f'{self._host}:{self._port}'
        if self._protocol:
            url = f'{self._protocol}://{url}'
        if self._resource:
            url = f'{url}/{self._resource}'
        return url


class LoadBalancing:
    def __init__(self, *sockets: str):
        """
        Class representing configuration of client-side load-balancing between multiple tick servers.
        Each client query selects the target tick server from the provided list according to the load-balancing policy.
        Currently only `SERVER_LOAD_WEIGHTED` policy is available -  tick servers are chosen at random using a weighted
        probability distribution, using inverted CPU usage for weights.
        LoadBalancing class is used only with :py:class:`~onetick.py.servers.RemoteTS` and
        :py:class:`~onetick.py.servers.FaultTolerance`.

        Parameters
        ----------
        sockets: str
            Sockets used in load-balancing configuration.

        Examples
        --------
        >>> LoadBalancing('host1:4001', 'host2:4002', 'host3:4003') # doctest: +SKIP
        """
        if len(sockets) < 2:
            raise ValueError(f'There must be at least 2 sockets for load balancing but {len(sockets)} was provided')
        self.sockets = sockets

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"({','.join(self.sockets)})"


class FaultTolerance:
    def __init__(self, *sockets: Union[str, 'LoadBalancing']):
        """
        Class representing configuration of client-side tick-server fault tolerance.
        Tick servers in the fault tolerance list are selected according to their level of priority,
        which is from left to right. Backup servers (servers after the first one, primary) are used only if the query
        on the primary server failed or server is down. So, if the first server is down the second server is used,
        if both of them are failed the third one used and so on.
        FaultTolerance class is used only with :py:class:`~onetick.py.servers.RemoteTS`.

        Parameters
        ----------
        sockets: str, LoadBalancing
            Sockets and/or LoadBalancing groups used in fault tolerance configuration.

        Examples
        --------
        >>> FaultTolerance('host1:4001', 'host2:4002', 'host3:4003') # doctest: +SKIP

        It is possible to have multiple load-balancing groups thus mixing together fault tolerance and load-balancing.
        The selection between different load balancing groups is also done according to their priority level,
        which is from left to right.

        >>> FaultTolerance(LoadBalancing('host1:4001', 'host2:4002'),
        ...                LoadBalancing('host3:4003', 'host3:4004')) # doctest: +SKIP
        """
        if len(sockets) < 2:
            raise ValueError(f'There must be at least 2 sockets for fault tolerance but {len(sockets)} was provided')
        self.sockets = sockets

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return ','.join(map(str, self.sockets))



import sys


def __search_main_one_tick_dir():
    try:
        import onetick.query_webapi  # noqa
        # if onetick.query_webapi is imported successfully, do not raise any warnings
        return __search_main_one_tick_dir_catch(show_warnings=False)
    except ImportError:
        return __search_main_one_tick_dir_catch(show_warnings=True)


def __search_main_one_tick_dir_catch(show_warnings=True):  # noqa: C901
    import os
    import warnings
    import sysconfig
    from pathlib import Path

    if os.environ.get('OTP_SKIP_OTQ_VALIDATION', False):
        try:
            import onetick_stubs  # noqa
        except ImportError:
            if show_warnings:
                warnings.warn(
                    "OTP_SKIP_OTQ_VALIDATION environment variable is set,"
                    " but onetick_stubs module is not available."
                    " Please, install onetick-query-stubs to avoid import errors"
                    " or unset OTP_SKIP_OTQ_VALIDATION to use onetick.query module."
                )
        return tuple()

    config_vars = sysconfig.get_config_vars()

    ot_bin_path = os.path.join("bin")
    ot_python_path = os.path.join(ot_bin_path, "python")
    ot_numpy_path = os.path.join(
        ot_bin_path,
        "numpy",
        "python"
        # short python version 27, 36, 37, etc
        + config_vars["py_version_nodot"]
        # suffix at the end, either empty string for python with the standard memory allocator,
        # or 'm' for python with the py-malloc allocator
        + config_vars["abiflags"],
    )

    if os.name == "nt":
        default_main_one_tick_dir = "C:/OMD/one_market_data/one_tick"
    elif os.name == "posix":
        default_main_one_tick_dir = "/opt/one_market_data/one_tick"
    else:
        default_main_one_tick_dir = None

    main_one_tick_dir = os.environ.get("MAIN_ONE_TICK_DIR")

    if not main_one_tick_dir:
        message = (
            "MAIN_ONE_TICK_DIR environment variable is not set."
            " It is a recommended way to let onetick-py know where OneTick python libraries are located."
        )
        if not default_main_one_tick_dir:
            if show_warnings:
                warnings.warn(message)
            return
        message += f" We will try to use default value for your system: {default_main_one_tick_dir}."
        if show_warnings:
            warnings.warn(message)
        main_one_tick_dir = default_main_one_tick_dir

    main_one_tick_dir = Path(main_one_tick_dir)

    if not main_one_tick_dir.is_dir():
        if show_warnings:
            warnings.warn(
                f"MAIN_ONE_TICK_DIR is set to '{main_one_tick_dir}',"
                " but this path is not a directory or doesn't exist."
            )
        return

    ot_bin_path = main_one_tick_dir / ot_bin_path
    ot_python_path = main_one_tick_dir / ot_python_path

    for directory in (ot_python_path, ot_bin_path):
        if not directory.is_dir():
            if show_warnings:
                warnings.warn(
                    f"MAIN_ONE_TICK_DIR is set to '{main_one_tick_dir}',"
                    f" and it must contain '{directory}' directory,"
                    " but this path is not a directory or doesn't exist."
                )
            return

    ot_numpy_path = main_one_tick_dir / ot_numpy_path

    return (ot_bin_path, ot_python_path, ot_numpy_path)


for directory in __search_main_one_tick_dir() or []:
    sys.path.append(str(directory))

__path__ = __import__("pkgutil").extend_path(__path__, __name__)



# This file is used to import the correct query module:
# onetick.query or onetick.query_webapi
# based on the environment variable OTP_WEBAPI
# Re-use this file in your tests to import the correct onetick.query module
# Also it is override pyomd and OneTickLib classes with mock classes

import getpass
import inspect
import os
import tempfile
import warnings
import onetick.py as otp


class OneTickLib:
    # mock class for OneTickLib, used for webapi and onetick-query-stubs
    # maybe it is not right to combine this namesake classes (onetick.lib.OneTickLib and pyomd.OneTickLib)
    # but it is done to simplify the code and it is not affecting anything except tests
    LOGGING_LEVEL_MIN = 0
    LOGGING_LEVEL_LOW = 1
    LOGGING_LEVEL_MEDIUM = 2
    LOGGING_LEVEL_MAX = 3
    set_authentication_token = None

    def __init__(self, *args, **kwargs):  # NOSONAR
        pass

    def set_log_file(self, log_file):  # NOSONAR
        pass

    def cleanup(self):  # NOSONAR
        pass


if os.getenv("OTP_SKIP_OTQ_VALIDATION"):
    import onetick_stubs as otq  # noqa: F401
    import pyomd  # noqa: F401

    class ConfigStub:
        API_CONFIG: dict = {}

    class otli:
        OneTickLib = OneTickLib  # NOSONAR

    for key, value in otq.query.__dict__.items():
        setattr(otq, key, value)
    setattr(otq, 'webapi', False)
    setattr(otq, 'config', ConfigStub)
    setattr(otq, 'graph_components', otq)

elif otp.__webapi__:
    import onetick.query_webapi as otq  # noqa: F401
    setattr(otq, 'webapi', True)

    from onetick.py.pyomd_mock import pyomd

    __original_run = otq.run

    def run(*args, **kwargs):
        from onetick.py import config  # noqa
        from onetick.py.compatibility import is_max_concurrency_with_webapi_supported

        if not config.http_address and 'http_address' not in kwargs:
            raise ValueError('otp.run() http_address keyword param, '
                             'otp.config.http_address or OTP_HTTP_ADDRESS '
                             'environment variable are required '
                             'when using WebAPI mode.')

        # TODO WEBAPI review this
        # if file name is not in single quotes, then put it in single quotes
        query_name = None
        query = kwargs.get('query', None)
        if isinstance(query, str):
            if query[0] == "'" and query[-1] == "'":
                query = query.replace("'", "")
            if '::' in query:
                query_name = query.split('::')[-1]
                query = query.replace(f'::{query_name}', '')
            kwargs['query'] = query
        kwargs['query_name'] = query_name

        del_params = [
            'start_time_expression',
            'end_time_expression',
            'alternative_username',
            'batch_size',
            'treat_byte_arrays_as_strings',
            'output_matrix_per_field',
            'return_utc_times',
            'connection',
            'svg_path',
            'use_connection_pool',
            'time_as_nsec',
            'max_expected_ticks_per_symbol',
        ]
        ignore_deleted_params = [
            'treat_byte_arrays_as_strings',
            'time_as_nsec',
            'alternative_username',
            'max_expected_ticks_per_symbol'
        ]
        for param in del_params:
            if param in kwargs:
                if kwargs[param] and param not in ignore_deleted_params:
                    warnings.warn(f'Parameter {param} is not supported in WebAPI mode and will be ignored.')
                del kwargs[param]

        from onetick.py import config  # noqa
        if 'http_address' not in kwargs:
            kwargs['http_address'] = config.http_address

        if kwargs.get('username'):
            kwargs['http_username'] = kwargs['username']
            del kwargs['username']
        else:
            kwargs['http_username'] = config.http_username

        if kwargs.get('password'):
            kwargs['http_password'] = kwargs['password']
            del kwargs['password']
        else:
            kwargs['http_password'] = config.http_password

        if 'access_token' not in kwargs:
            if config.access_token:
                kwargs['access_token'] = config.access_token
        elif not kwargs['access_token']:
            del kwargs['access_token']

        access_token_url = kwargs.get('access_token_url', config.access_token_url)
        if access_token_url:
            if not hasattr(otq, 'get_access_token'):
                raise RuntimeError('Current `onetick.query_webapi` version doesn\'t have `get_access_token` function')

            if kwargs.get('access_token'):
                raise ValueError('Both `access_token` and `access_token_url` set, instead of only one of them.')

            if 'client_id' in kwargs:
                client_id = kwargs.pop('client_id')
            else:
                client_id = config.client_id

            if 'client_secret' in kwargs:
                client_secret = kwargs.pop('client_secret')
            else:
                client_secret = config.client_secret

            for param_name, param_value in zip(['client_id', 'client_secret'], [client_id, client_secret]):
                if not param_value:
                    raise ValueError(f'`access_token_url` parameter set, however `{param_name}` parameter missing.')

            kwargs['access_token'] = otq.get_access_token(access_token_url, client_id, client_secret)

            if 'access_token_url' in kwargs:
                del kwargs['access_token_url']

        if 'http_proxy' not in kwargs:
            kwargs['http_proxy'] = config.http_proxy

        if 'https_proxy' not in kwargs:
            kwargs['https_proxy'] = config.https_proxy

        is_trusted_certificates_supported = 'trusted_certificates_file' in inspect.signature(otq.run).parameters

        if (
            ('trusted_certificates_file' in kwargs or config.trusted_certificates_file is not None) and
            not is_trusted_certificates_supported
        ):
            raise ValueError(
                'Parameter `trusted_certificates_file` was set, however current version of OneTick doesn\'t support it.'
            )

        if 'trusted_certificates_file' not in kwargs and is_trusted_certificates_supported:
            kwargs['trusted_certificates_file'] = config.trusted_certificates_file

        if 'callback' in kwargs and kwargs['callback'] is not None:
            kwargs['output_mode'] = otq.QueryOutputMode.callback

        if 'max_concurrency' in kwargs and not is_max_concurrency_with_webapi_supported():
            kwargs['max_concurrency'] = None

        return __original_run(*args, **kwargs)

    otq.run = run

    otq.OneTickLib = OneTickLib

    class otli:  # type: ignore # noqa: F401
        OneTickLib = otq.OneTickLib  # NOSONAR

else:
    import onetick.query as otq  # type: ignore # noqa: F401
    import pyomd  # type: ignore # noqa: F401
    import onetick.lib.instance as otli  # type: ignore # noqa: F401
    setattr(otq, 'webapi', False)


def _tmp_otq_path():
    # copied from onetick.test.fixtures _keep_generated_dir() with replacement to /tmp
    # required with OTP_WEBAPI_TEST_MODE to separate otq files from shared dbs+config+locator+acl files
    from onetick.py import utils  # noqa
    res = os.path.join(utils.TMP_CONFIGS_DIR(), os.environ.get("ONE_TICK_TMP_DIR", "otqs"))
    res = res.replace(utils.temp.WEBAPI_TEST_MODE_SHARED_CONFIG,
                      os.path.join(tempfile.gettempdir(), "test_" + getpass.getuser()))

    # % and + is some webapi related path bugs, probably would be fixed someday
    return res.replace("%", "_").replace("+", "_")


__all__ = ['otq', 'pyomd', 'otli', '_tmp_otq_path']



import os
import sys
import multiprocessing
import warnings


def get_local_number_of_cores():
    try:
        # number of cores allowed for user
        return len(os.sched_getaffinity(0))
    except Exception:
        # number of cores
        return multiprocessing.cpu_count()


def default_license_dir():
    if sys.platform == "linux":
        return os.path.join("/", "license")
    elif sys.platform == "win32":
        return os.path.join("C:\\", "OMD", "client_data", "config", "license_repository")
    elif sys.platform == "darwin":
        return os.path.join("/", "Library", "Application Support", "OneTick")
    return None


def default_license_file():
    if sys.platform == "linux":
        return os.path.join("/", "license", "license.dat")
    elif sys.platform == "win32":
        return os.path.join("C:\\", "OMD", "client_data", "config", "license.dat")
    elif sys.platform == "darwin":
        return os.path.join("/", "Library", "Application Support", "OneTick")
    return None


def default_day_boundary_tz(db_name):
    import onetick.py as otp
    if otp.config.tz:
        return otp.config.tz
    warnings.warn(
        "Database property 'day_boundary_tz' can't be set to default timezone "
        "because default timezone is local and is not known at this moment. "
        "Default value for this property will be 'GMT'. "
        f"It may produce unexpected results "
        f"when reading from or writing to database '{db_name}'. "
        "Set otp.config.tz property to some known value to avoid this situation.",
        stacklevel=2,
    )
    return None



from typing import Union, TYPE_CHECKING
from copy import deepcopy

if TYPE_CHECKING:
    from onetick.py.core.source import Source   # hack for annotations

from onetick.py.core.column import _Column
from onetick.py import types as ott
from onetick.py.otq import otq

from ._base import (
    _AggregationTSType, _AggregationTSSelection, _KeepTs, _FloatAggregation, _ExpectLargeInts, _AllColumnsAggregation,
)


class Max(_AggregationTSType, _ExpectLargeInts):

    NAME = "HIGH"
    EP = otq.High
    require_type = (int, float, ott.nsectime, ott._inf)

    FIELDS_MAPPING = deepcopy(_AggregationTSType.FIELDS_MAPPING)
    FIELDS_MAPPING.update(_ExpectLargeInts.FIELDS_MAPPING)
    FIELDS_DEFAULT = deepcopy(_AggregationTSType.FIELDS_DEFAULT)
    FIELDS_DEFAULT.update(_ExpectLargeInts.FIELDS_DEFAULT)


class Min(Max):
    NAME = "LOW"
    EP = otq.Low


class HighTick(_AggregationTSType, _AggregationTSSelection, _KeepTs, _FloatAggregation, _AllColumnsAggregation):
    EP = otq.HighTick
    NAME = 'HIGH_TICK'
    DEFAULT_OUTPUT_NAME = 'HIGH_TICK'

    FIELDS_MAPPING = deepcopy(_AggregationTSType.FIELDS_MAPPING)
    FIELDS_MAPPING.update(_AggregationTSSelection.FIELDS_MAPPING)
    FIELDS_MAPPING['n'] = 'NUM_TICKS'
    FIELDS_DEFAULT = deepcopy(_AggregationTSType.FIELDS_DEFAULT)
    FIELDS_DEFAULT.update(_AggregationTSSelection.FIELDS_DEFAULT)
    FIELDS_DEFAULT['n'] = 1

    FIELDS_TO_SKIP = ['output_field_name', 'all_fields']

    def __init__(self, column: Union[str, _Column], n: int = 1, *args, **kwargs):
        """
        Select `n` ticks with the highest values in the `column` field
        """
        super().__init__(column, *args, **kwargs)
        self.n = n

    @staticmethod
    def validate_output_name(*args, **kwargs):
        # HighTick and LowTick aggregations don't have output fields
        pass


class LowTick(HighTick):
    EP = otq.LowTick
    NAME = 'LOW_TICK'
    DEFAULT_OUTPUT_NAME = 'LOW_TICK'


class HighTime(_AggregationTSType, _AggregationTSSelection, _FloatAggregation):
    NAME = "HIGH_TIME"
    EP = otq.HighTime

    FIELDS_MAPPING = deepcopy(_AggregationTSType.FIELDS_MAPPING)
    FIELDS_MAPPING.update(_AggregationTSSelection.FIELDS_MAPPING)
    FIELDS_DEFAULT = deepcopy(_AggregationTSType.FIELDS_DEFAULT)
    FIELDS_DEFAULT.update(_AggregationTSSelection.FIELDS_DEFAULT)
    output_field_type = ott.nsectime


class LowTime(HighTime):
    """Returns timestamp of tick with lowest value of input field"""
    NAME = "LOW_TIME"
    EP = otq.LowTime



from onetick.py import types as ott
from onetick.py.core.column_operations.accessors._accessor import _Accessor


class _FloatAccessor(_Accessor):
    """
    Accessor for float (double in Onetick terminology) functions

    >>> data = otp.Ticks(X=[1.1, 1.2])
    >>> data["Y"] = data["X"].float.<function_name>()   # doctest: +SKIP
    """

    def str(self, length=10, precision=6):
        """
        Converts float to str.

        Converts number to string with given ``length`` and ``precision``.
        The specified ``length`` should be greater than or equal
        to the part of the number before the decimal point plus the number's sign (if any).

        If ``length`` is specified as an int, the method will return strings with ``length`` characters,
        if ``length`` is specified as a column, the method will return string default (64 characters) length.

        Parameters
        ----------
        length: Operation or int
            Length of the string.
        precision: Operation or int
            Number of symbols after dot.

        Returns
        -------
        result: Operation
            String representation of float value.

        Examples
        --------

        >>> data = otp.Ticks(X=[1, 2.17, 10.31861, 3.141593])
        >>> # OTdirective: snippet-name: float operations.to string.constant precision;
        >>> data["X"] = data["X"].float.str(15, 3)
        >>> data = otp.run(data)
        >>> data["X"]
        0    1.000
        1    2.170
        2    10.319
        3    3.142
        Name: X, dtype: object

        >>> data = otp.Ticks(X=[1, 2.17, 10.31841, 3.141593],
        ...                  LENGTH=[2, 3, 4, 5],
        ...                  PRECISION=[5, 5, 3, 3])
        >>> # OTdirective: snippet-name: float operations.to string.precision from fields;
        >>> data["X"] = data["X"].float.str(data["LENGTH"], data["PRECISION"])
        >>> data = otp.run(data)
        >>> data["X"]
        0    1
        1    2.2
        2    10.3
        3    3.142
        Name: X, dtype: object
        """
        dtype = ott.string[length] if isinstance(length, int) else str
        length = ott.value2str(length)
        precision = ott.value2str(precision)
        return _FloatAccessor.Formatter(self._base_column,
                                        dtype,
                                        formatter=lambda x: f"str({x}, {length}, {precision})")

    def cmp(self, other, eps):
        """
        Compare two double values between themselves according to ``eps`` relative difference.

        This function returns 0 if column = other, 1 if column > other, and -1 if column < other.
        Two numbers are considered to be equal if both of them are NaN or
        ``abs(column - other) / (abs(column) + abs(other)) < eps``.
        In other words, ``eps`` represents a relative difference (percentage) between the two numbers,
        not an absolute difference.

        Parameters
        ----------
        other: Operation or float
            column or value to compare with
        eps: Operation or float
            column or value with relative difference

        Returns
        -------
        result: Operation
            0 if column == other, 1 if column > other, and -1 if column < other.

        See Also
        --------
        eq

        Examples
        --------

        >>> data = otp.Ticks(X=[1, 2.17, 10.31841, 3.141593, 6],
        ...                  OTHER=[1.01, 2.1, 10.32841, 3.14, 5],
        ...                  EPS=[0, 1, 0.1, 0.001, 0.001])
        >>> # OTdirective: snippet-name: float operations.approximate comparison.lt|eq|gt;
        >>> data["X"] = data["X"].float.cmp(data["OTHER"], data["EPS"])
        >>> data = otp.run(data)
        >>> data["X"]
        0   -1.0
        1    0.0
        2    0.0
        3    0.0
        4    1.0
        Name: X, dtype: float64
        """
        other = ott.value2str(other)
        eps = ott.value2str(eps)
        return _FloatAccessor.Formatter(self._base_column,
                                        float,
                                        formatter=lambda x: f"double_compare({x}, {other}, {eps})")

    def eq(self, other, delta):
        """
        Compare two double values between themselves according to ``delta`` relative difference.
        Calculated as ``abs(column - other) <= delta``.

        Parameters
        ----------
        other: Operation, float
            column or value to compare with
        delta: Operation, float
            column or value with relative difference

        See Also
        --------
        cmp

        Returns
        -------
            Operation

        Examples
        --------

        >>> data = otp.Ticks(X=[1, 2.17, 10.31841, 3.141593, 6],
        ...                  OTHER=[1.01, 2.1, 10.32841, 3.14, 5],
        ...                  DELTA=[0, 1, 0.1, 0.01, 0.001])
        >>> # OTdirective: snippet-name: float operations.approximate comparison.eq;
        >>> data["X"] = (1 + data["X"] - 1).float.eq(data["OTHER"], data["DELTA"])
        >>> data = otp.run(data)
        >>> data["X"]
        0    0.0
        1    1.0
        2    1.0
        3    1.0
        4    0.0
        Name: X, dtype: float64
        """
        other = ott.value2str(other)
        delta = ott.value2str(delta)
        return _FloatAccessor.Formatter(self._base_column,
                                        bool,
                                        formatter=lambda x: f"abs({x} - {other}) <= {delta}")



from .methods import fillna, isin, round, _map
from .methods import add, sub, mul, mod, div, abs, neg, pos
from .methods import eq, ne, and_, or_, gt, ge, lt, le, invert
from .conversions import CONVERSIONS



from typing import Iterable, Callable

import onetick.py as otp
from onetick.py.docs.utils import docstring

from ..aggregations.order_book import (
    OB_SNAPSHOT_DOC_PARAMS,
    OB_SNAPSHOT_WIDE_DOC_PARAMS,
    OB_SNAPSHOT_FLAT_DOC_PARAMS,
    OB_SUMMARY_DOC_PARAMS,
    OB_SIZE_DOC_PARAMS,
    OB_VWAP_DOC_PARAMS,
    OB_NUM_LEVELS_DOC_PARAMS,
)
from ..aggregations.functions import (
    ob_snapshot, ob_snapshot_wide, ob_snapshot_flat, ob_summary, ob_size, ob_vwap, ob_num_levels,
)

from .data_source import DataSource, DATA_SOURCE_DOC_PARAMS


class _ObSource(DataSource):
    OB_AGG_FUNC: Callable
    OB_AGG_PARAMS: Iterable
    _PROPERTIES = DataSource._PROPERTIES + ['_ob_agg']

    def __init__(self, *args, schema=None, **kwargs):
        if self._try_default_constructor(*args, schema=schema, **kwargs):
            return

        ob_agg_params = {
            param.name: kwargs.pop(param.name, param.default)
            for _, param in self.OB_AGG_PARAMS
        }
        self._ob_agg = self.__class__.OB_AGG_FUNC(**ob_agg_params)

        super().__init__(*args, schema=schema, **kwargs)

        ob_agg_output_schema = self._ob_agg._get_output_schema(otp.Empty())

        if getattr(self._ob_agg, 'show_full_detail', None):
            self.schema.update(**ob_agg_output_schema)
        else:
            self.schema.set(**ob_agg_output_schema)

    def base_ep(self, *args, **kwargs):
        src = super().base_ep(*args, **kwargs)
        return self._ob_agg.apply(src)

    def _base_ep_for_cross_symbol(self, *args, **kwargs):
        src = super()._base_ep_for_cross_symbol(*args, **kwargs)
        return self._ob_agg.apply(src)


@docstring(parameters=OB_SNAPSHOT_DOC_PARAMS + DATA_SOURCE_DOC_PARAMS)
class ObSnapshot(_ObSource):
    """
    Construct a source providing order book snapshot for a given ``db``.
    This is just a shortcut for
    :class:`~onetick.py.DataSource` + :func:`~onetick.py.agg.ob_snapshot`.

    See also
    --------
    | :class:`onetick.py.DataSource`
    | :meth:`onetick.py.Source.ob_snapshot`
    | :func:`onetick.py.agg.ob_snapshot`
    | **OB_SNAPSHOT** OneTick event processor

    Examples
    ---------

    >>> data = otp.ObSnapshot(db='SOME_DB', tick_type='PRL', symbols='AA', max_levels=1) # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  PRICE             UPDATE_TIME  SIZE  LEVEL  BUY_SELL_FLAG
    0 2003-12-04    2.0 2003-12-01 00:00:00.003     6      1              1
    1 2003-12-04    5.0 2003-12-01 00:00:00.004     7      1              0
    """
    OB_AGG_FUNC = ob_snapshot
    OB_AGG_PARAMS = OB_SNAPSHOT_DOC_PARAMS


@docstring(parameters=OB_SNAPSHOT_WIDE_DOC_PARAMS + DATA_SOURCE_DOC_PARAMS)
class ObSnapshotWide(_ObSource):
    """
    Construct a source providing order book wide snapshot for a given ``db``.
    This is just a shortcut for
    :class:`~onetick.py.DataSource` + :func:`~onetick.py.agg.ob_snapshot_wide`.

    See also
    --------
    | :class:`onetick.py.DataSource`
    | :meth:`onetick.py.Source.ob_snapshot_wide`
    | :func:`onetick.py.agg.ob_snapshot_wide`
    | **OB_SNAPSHOT_WIDE** OneTick event processor

    Examples
    ---------

    >>> data = otp.ObSnapshotWide(db='SOME_DB', tick_type='PRL', symbols='AA', max_levels=1) # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  BID_PRICE         BID_UPDATE_TIME  BID_SIZE  ASK_PRICE         ASK_UPDATE_TIME  ASK_SIZE  LEVEL
    0 2003-12-03        5.0 2003-12-01 00:00:00.004         7        2.0 2003-12-01 00:00:00.003         6      1
    """
    OB_AGG_FUNC = ob_snapshot_wide
    OB_AGG_PARAMS = OB_SNAPSHOT_WIDE_DOC_PARAMS


@docstring(parameters=OB_SNAPSHOT_FLAT_DOC_PARAMS + DATA_SOURCE_DOC_PARAMS)
class ObSnapshotFlat(_ObSource):
    """
    Construct a source providing order book flat snapshot for a given ``db``.
    This is just a shortcut for
    :class:`~onetick.py.DataSource` + :func:`~onetick.py.agg.ob_snapshot_flat`.

    See also
    --------
    | :class:`onetick.py.DataSource`
    | :meth:`onetick.py.Source.ob_snapshot_flat`
    | :func:`onetick.py.agg.ob_snapshot_flat`
    | **OB_SNAPSHOT_FLAT** OneTick event processor

    Examples
    ---------

    >>> data = otp.ObSnapshotFlat(db='SOME_DB', tick_type='PRL', symbols='AA', max_levels=1) # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  BID_PRICE1        BID_UPDATE_TIME1  BID_SIZE1  ASK_PRICE1        ASK_UPDATE_TIME1  ASK_SIZE1
    0 2003-12-03         5.0 2003-12-01 00:00:00.004          7         2.0 2003-12-01 00:00:00.003          6
    """
    OB_AGG_FUNC = ob_snapshot_flat
    OB_AGG_PARAMS = OB_SNAPSHOT_FLAT_DOC_PARAMS


@docstring(parameters=OB_SUMMARY_DOC_PARAMS + DATA_SOURCE_DOC_PARAMS)
class ObSummary(_ObSource):
    """
    Construct a source providing order book summary for a given ``db``.
    This is just a shortcut for
    :class:`~onetick.py.DataSource` + :func:`~onetick.py.agg.ob_summary`.

    See also
    --------
    | :class:`onetick.py.DataSource`
    | :meth:`onetick.py.Source.ob_summary`
    | :func:`onetick.py.agg.ob_summary`
    | **OB_SUMMARY** OneTick event processor

    Examples
    ---------

    >>> data = otp.ObSummary(db='SOME_DB', tick_type='PRL', symbols='AA', max_levels=1) # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  BID_PRICE  BID_SIZE  BID_VWAP  BEST_BID_PRICE  WORST_BID_SIZE  NUM_BID_LEVELS  ASK_SIZE\
              ASK_VWAP  BEST_ASK_PRICE  WORST_ASK_PRICE  NUM_ASK_LEVELS
    0 2003-12-04        NaN         7       5.0             5.0             NaN               1         6\
           2.0             2.0              2.0               1
    """
    OB_AGG_FUNC = ob_summary
    OB_AGG_PARAMS = OB_SUMMARY_DOC_PARAMS


@docstring(parameters=OB_SIZE_DOC_PARAMS + DATA_SOURCE_DOC_PARAMS)
class ObSize(_ObSource):
    """
    Construct a source providing number of order book levels for a given ``db``.
    This is just a shortcut for
    :class:`~onetick.py.DataSource` + :func:`~onetick.py.agg.ob_size`.

    See also
    --------
    | :class:`onetick.py.DataSource`
    | :meth:`onetick.py.Source.ob_size`
    | :func:`onetick.py.agg.ob_size`
    | **OB_SIZE** OneTick event processor

    Examples
    ---------

    >>> data = otp.ObSize(db='SOME_DB', tick_type='PRL', symbols='AA', max_levels=10) # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  ASK_VALUE  BID_VALUE
    0 2003-12-01      84800      64500
    """
    OB_AGG_FUNC = ob_size
    OB_AGG_PARAMS = OB_SIZE_DOC_PARAMS


@docstring(parameters=OB_VWAP_DOC_PARAMS + DATA_SOURCE_DOC_PARAMS)
class ObVwap(_ObSource):
    """
    Construct a source providing the size-weighted price
    computed over a specified number of order book levels for a given ``db``.
    This is just a shortcut for
    :class:`~onetick.py.DataSource` + :func:`~onetick.py.agg.ob_vwap`.

    See also
    --------
    | :class:`onetick.py.DataSource`
    | :meth:`onetick.py.Source.ob_vwap`
    | :func:`onetick.py.agg.ob_vwap`
    | **OB_VWAP** OneTick event processor

    Examples
    ---------

    >>> data = otp.ObVwap(db='SOME_DB', tick_type='PRL', symbols='AA', max_levels=10) # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  ASK_VALUE  BID_VALUE
    0 2003-12-01     23.313   23.20848
    """
    OB_AGG_FUNC = ob_vwap
    OB_AGG_PARAMS = OB_VWAP_DOC_PARAMS


@docstring(parameters=OB_NUM_LEVELS_DOC_PARAMS + DATA_SOURCE_DOC_PARAMS)
class ObNumLevels(_ObSource):
    """
    Construct a source providing the number of levels in the order book for a given ``db``.
    This is just a shortcut for
    :class:`~onetick.py.DataSource` + :func:`~onetick.py.agg.ob_num_levels`.

    See also
    --------
    | :class:`onetick.py.DataSource`
    | :meth:`onetick.py.Source.ob_num_levels`
    | :func:`onetick.py.agg.ob_num_levels`
    | **OB_NUM_LEVELS** OneTick event processor

    Examples
    ---------

    >>> data = otp.ObNumLevels(db='SOME_DB', tick_type='PRL', symbols='AA') # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  ASK_VALUE  BID_VALUE
    0 2003-12-01        248         67
    """
    OB_AGG_FUNC = ob_num_levels
    OB_AGG_PARAMS = OB_NUM_LEVELS_DOC_PARAMS
